import statistics

from fontTools import unicodedata

from openpecha.config import get_logger
from openpecha.pecha.parsers.ocr.ocr import UNICODE_CHARCAT_FOR_WIDTH, BBox, OCRParser

# Initialize the logger
logger = get_logger(__name__)


class GoogleVisionParser(OCRParser):
    """
    OpenPecha Parser for Google OCR JSON output of scanned pecha.
    """

    def __init__(self, output_path=None, metadata=None):
        super().__init__(output_path, metadata)
        self.check_postprocessing = False

    def has_space_attached(self, symbol):
        """Checks if symbol has space followed by it or not

        Args:
            symbol (dict): symbol info

        Returns:
            boolean: True if the symbol has space followed by it
        """
        if (
            "property" in symbol
            and "detectedBreak" in symbol["property"]
            and "type" in symbol["property"]["detectedBreak"]
            and symbol["property"]["detectedBreak"]["type"] == "SPACE"
        ):
            return True
        return False

    def get_language_code_from_gv_poly(self, gv_poly):
        lang = ""
        properties = gv_poly.get("property", {})
        if properties:
            languages = properties.get("detectedLanguages", [])
            if languages:
                lang = languages[0]["languageCode"]
        if lang == "" or lang == "und":
            # this is not always true but replacing it with None is worse
            # with our current data
            return self.default_language
        if lang in ["bo", "en", "zh"]:
            return lang
        if lang == "dz":
            return "bo"
        # English is a kind of default for our purpose
        return "en"

    def get_bboxinfo_from_vertices(self, vertices):
        """
        Vertices do not always have dots in the same order. The current hypothesis
        is that their order represents the rotation of characters detected by
        the OCR.

        This is not documented on

        https://cloud.google.com/vision/docs/reference/rest/v1/projects.locations.products.referenceImages#Vertex

        though, so use the angle value with caution.
        """
        if len(vertices) == 0:
            return None
        idx_smallest = -1
        smallest_x = -1
        smallest_y = -1
        largest_x = -1
        largest_y = -1
        for idx, v in enumerate(vertices):
            if "x" not in v or "y" not in v:
                continue
            smallest_x = v["x"] if smallest_x == -1 else min(v["x"], smallest_x)
            smallest_y = v["y"] if smallest_y == -1 else min(v["y"], smallest_y)
            largest_x = max(v["x"], largest_x)
            largest_y = max(v["y"], largest_y)
            # here we have to account for cases where the 4 dots don't form a rectangle
            # because coordinates are shifted by 1, see test_bbox_info for some example
            if abs(v["x"] - smallest_x) < 3 and abs(v["y"] - smallest_y) < 3:
                idx_smallest = idx
        if smallest_x == -1 or smallest_y == -1 or largest_y == -1 or largest_x == -1:
            return None
        angle = None
        if len(vertices) == 4 and idx_smallest != -1:
            angle = 0
            if idx_smallest == 1:
                angle = 270
            if idx_smallest == 2:
                angle = 180
            if idx_smallest == 3:
                angle = 90
        return [smallest_x, largest_x, smallest_y, largest_y, angle]

    def dict_to_bbox(self, word):
        """Convert bounding bbox to BBox object

        Args:
            word (dict): bounding gv_poly of a word infos

        Returns:
            obj: BBox object of bounding bbox
        """
        confidence = word.get("confidence")
        if "boundingBox" not in word or "vertices" not in word["boundingBox"]:
            return None
        vertices = word["boundingBox"]["vertices"]
        bboxinfo = self.get_bboxinfo_from_vertices(vertices)
        if bboxinfo is None:
            return None
        if self.remove_rotated_boxes and bboxinfo[4] > 0:
            return None
        return BBox(
            bboxinfo[0],
            bboxinfo[1],
            bboxinfo[2],
            bboxinfo[3],
            bboxinfo[4],
            confidence=confidence,
        )

    def get_width_of_vertices(self, vertices):
        if len(vertices) < 4:
            return None
        # oddly enough, sometimes Google returns a vertex with x=-1...
        smallest_x = None
        largest_x = -1
        for v in vertices:
            if "x" not in v or "y" not in v:
                continue
            smallest_x = v["x"] if smallest_x is None else min(v["x"], smallest_x)
            largest_x = max(v["x"], largest_x)
        if smallest_x is None:
            return None
        dif = largest_x - smallest_x
        return dif

    def get_char_base_bboxes_and_avg_width(self, response):
        """Return bounding bboxs in page response

        Args:
            response (dict): google ocr output of a page

        Returns:
            list: list of BBox object which saves required info of a bounding bbox
        """
        bboxes = []
        widths = []
        for page in response["fullTextAnnotation"]["pages"]:
            for block in page["blocks"]:
                cur_line_boxes = []
                for paragraph in block["paragraphs"]:
                    for word in paragraph["words"]:
                        bbox = self.dict_to_bbox(word)
                        if bbox is None:
                            # case where we ignore the bbox for some reason
                            # for instance rotated text
                            continue
                        cur_word = ""
                        for symbol in word["symbols"]:
                            symbolunicat = unicodedata.category(symbol["text"][0])
                            if symbolunicat in UNICODE_CHARCAT_FOR_WIDTH:
                                vertices = symbol["boundingBox"]["vertices"]
                                width = self.get_width_of_vertices(vertices)
                                if width is not None and width > 0:
                                    widths.append(width)
                            cur_word += symbol["text"]
                            if self.has_space_attached(symbol):
                                cur_word += " "
                        if cur_word:
                            bbox.text = cur_word
                            # the language returned by Google OCR is not particularly helpful
                            # language = self.get_language_code_from_gv_poly(word)
                            # instead we use our custom detection system
                            bbox.language = self.get_main_language_code(cur_word)
                            cur_line_boxes.append(bbox)
                bboxes.append(cur_line_boxes)
        avg_width = statistics.mean(widths) if widths else None
        logger.debug("average char width: %f", avg_width)
        return bboxes, avg_width

    def get_bboxes_for_page(self, image_group_id, image_filename):
        ocr_object = self.data_source.get_image_data(image_group_id, image_filename)
        try:
            page_content = ocr_object["textAnnotations"][0]["description"]
        except Exception as e:
            logger.error(
                f"OCR page is empty {page_content} (no textAnnotations[0]/description): {e}"
            )
            return None, 0
        return self.get_char_base_bboxes_and_avg_width(ocr_object)



# import re
# from pathlib import Path
# from typing import Any, Dict

# from openpecha.config import PECHAS_PATH
# from openpecha.pecha import Pecha
# from openpecha.pecha.layer import AnnotationType
# from openpecha.pecha.parsers import BaseParser


# class plaintextparser(BaseParser):
#     def __init__(self):
#         self.temp_state = {
#             "base_text": "",
#             "annotations": {
#                 "segments": {},
#                 "pages": {},
#             },
#             "prev_info_dict": {},
#         }

#     def extract_segment_id_and_line(self, text):
#         pattern = r"\[(\d+[ab](?:\.\d+)?)\]"
#         results = []

#         matches = list(re.finditer(pattern, text))

#         if not matches:
#             results.append({"page_number": None, "line": text.strip()})
#             return results

#         if matches[0].start() > 0:
#             pre_content = text[: matches[0].start()].strip()
#             if pre_content:
#                 results.append({"page_number": None, "line": pre_content})

#         for i, match in enumerate(matches):
#             page_number = match.group(1)
#             start = match.end()
#             end = matches[i + 1].start() if i + 1 < len(matches) else len(text)
#             content = text[start:end].strip()

#             if not content:
#                 results.append({"page_number": page_number, "line": ""})
#                 continue

#             lines = content.splitlines()
#             for line in lines:
#                 line = line.strip()
#                 line = re.sub(r"^\{.*?\}#?", "", line)
#                 if line:
#                     results.append({"page_number": page_number, "line": line})

#         return results

#     def get_page_id(self, segment_id):
#         if "." in segment_id:
#             return segment_id.split(".")[0]
#         else:
#             return segment_id

#     def get_info_dict(self, segment_id, page_id, text):
#         text_len = len(text)
#         curr_dict: Dict[str, Any] = {"annotations": {"segments": {}, "pages": {}}}
#         if self.temp_state["prev_info_dict"]:
#             segment_start = (
#                 self.temp_state["prev_info_dict"]["segments"]["span"]["end"] + 1
#             )
#             segment_end = segment_start + text_len
#             curr_dict["annotations"]["segments"][segment_id] = {
#                 "span": {"start": segment_start, "end": segment_end}
#             }
#             if page_id == self.temp_state["prev_info_dict"]["page_id"]:
#                 page_start = self.temp_state["prev_info_dict"]["pages"]["span"]["start"]
#                 curr_dict["annotations"]["pages"][page_id] = {
#                     "span": {"start": page_start, "end": segment_end}
#                 }
#             elif page_id != self.temp_state["prev_info_dict"]["page_id"]:
#                 page_start = (
#                     self.temp_state["prev_info_dict"]["pages"]["span"]["end"] + 1
#                 )
#                 curr_dict["annotations"]["pages"][page_id] = {
#                     "span": {"start": page_start, "end": page_start + text_len}
#                 }
#         else:
#             curr_dict["annotations"]["segments"][segment_id] = {
#                 "span": {"start": 0, "end": text_len}
#             }
#             curr_dict["annotations"]["pages"][page_id] = {
#                 "span": {"start": 0, "end": text_len}
#             }
#         return curr_dict

#     def parse_bo(self, text):
#         text_list = text.split("\n")
#         for segment in text_list:
#             result = self.extract_segment_id_and_line(segment)
#             for line in result:
#                 segment_id = line["page_number"]
#                 line_text = line["line"]
#                 if not segment_id or "." not in segment_id:
#                     continue
#                 page_id = self.get_page_id(segment_id)
#                 curr_dict = self.get_info_dict(segment_id, page_id, line_text)
#                 if self.temp_state["base_text"] == "":
#                     self.temp_state["base_text"] = line_text
#                 elif (
#                     self.temp_state["base_text"] != ""
#                     and self.temp_state["prev_info_dict"]["page_id"] == page_id
#                 ):
#                     self.temp_state["base_text"] += "\n" + line_text
#                 elif self.temp_state["prev_info_dict"]["page_id"] != page_id:
#                     self.temp_state["base_text"] += "\n" + line_text
#                 self.temp_state["annotations"]["segments"][segment_id] = curr_dict[
#                     "annotations"
#                 ]["segments"][segment_id]
#                 self.temp_state["annotations"]["pages"][page_id] = curr_dict[
#                     "annotations"
#                 ]["pages"][page_id]
#                 self.temp_state["prev_info_dict"] = {
#                     "segment_id": segment_id,
#                     "page_id": page_id,
#                     "segments": curr_dict["annotations"]["segments"][segment_id],
#                     "pages": curr_dict["annotations"]["pages"][page_id],
#                 }

#     def get_info_dict_for_zh(self, segment_id, segment):
#         text_len = len(segment)
#         if text_len == 0:
#             segment = "\n"
#             text_len = 1
#         curr_dict = {"annotations": {"segments": {}, "pages": {}}}
#         if self.temp_state["base_text"] == "":
#             segment_start = 0
#             segment_end = text_len
#             self.temp_state["base_text"] = segment + "\n"
#             self.temp_state["annotations"]["segments"][segment_id] = {
#                 "span": {"start": segment_start, "end": segment_end}
#             }
#         else:
#             segment_start = len(self.temp_state["base_text"])
#             segment_end = segment_start + text_len
#             if segment == "\n":
#                 self.temp_state["base_text"] += segment
#             else:
#                 self.temp_state["base_text"] += segment + "\n"
#             self.temp_state["annotations"]["segments"][segment_id] = {
#                 "span": {"start": segment_start, "end": segment_end}
#             }
#         curr_dict["annotations"]["segments"][segment_id] = {
#             "span": {"start": segment_start, "end": segment_end}
#         }
#         return curr_dict

#     def parse_zh(self, text):
#         text_list = text.split("\n")
#         for segment_id, segment in enumerate(text_list, 1):
#             curr_dict = self.get_info_dict_for_zh(segment_id, segment)
#             self.temp_state["prev_info_dict"] = {
#                 "segment_id": segment_id,
#                 "segments": curr_dict["annotations"]["segments"][segment_id],
#             }

#     def write_to_pecha(self, pecha, metadata):
#         curr_bases = {}
#         bases = []
#         order = 1
#         base_name = pecha.set_base(content=self.temp_state["base_text"])

#         total_segments = 0
#         total_pages = 0

#         if "segments" in self.temp_state["annotations"]:
#             segment, _ = pecha.add_layer(base_name, AnnotationType.segmentation)
#             for segment_id, segment_span in self.temp_state["annotations"][
#                 "segments"
#             ].items():
#                 segment_ann = {
#                     AnnotationType.segmentation.value: segment_span["span"],
#                     "segment_id": segment_id,
#                 }
#                 pecha.add_annotation(segment, segment_ann, AnnotationType.segmentation)
#             segment.save()
#             total_segments = len(self.temp_state["annotations"]["segments"])

#         if "pages" in self.temp_state["annotations"]:
#             pagination, _ = pecha.add_layer(base_name, AnnotationType.pagination)
#             for page_id, page_ann in self.temp_state["annotations"]["pages"].items():
#                 page_ann = {
#                     AnnotationType.pagination.value: page_ann["span"],
#                     "folio": page_id,
#                 }
#                 pecha.add_annotation(pagination, page_ann, AnnotationType.pagination)
#             pagination.save()
#             total_pages = len(self.temp_state["annotations"]["pages"])

#         curr_bases = {
#             base_name: {
#                 "source_metadata": {
#                     "source_id": "001",
#                     "total_segments": total_segments,
#                     "total_pages": total_pages,
#                 },
#                 "base_file": base_name,
#                 "order": order,
#             }
#         }
#         bases.append(curr_bases)

#         pecha.set_metadata(
#             {"id": pecha.id, "parser": self.name, "bases": bases, **metadata}
#         )

#     def parse(
#         self,
#         input: Any,
#         metadata: Dict|Path,
#         output_path: Path = PECHAS_PATH,
#     ):
#         text = input.read_text(encoding="utf-8")
#         if metadata["language"] == "bo":
#             self.parse_bo(text)
#         elif metadata["language"] == "zh":
#             self.parse_zh(text)
#         pecha = Pecha.create(output_path)
#         self.write_to_pecha(pecha, metadata)
#         return pecha



import ast
from pathlib import Path
from typing import List

from git import Repo

from openpecha.config import GITHUB_ORG_NAME, PECHAS_PATH
from openpecha.github_utils import clone_repo
from openpecha.storages import commit_and_push
from openpecha.utils import read_csv, write_csv


class PechaDataCatalog:
    def __init__(self, output_path: Path = PECHAS_PATH):
        self.org_name = GITHUB_ORG_NAME
        self.repo_name = "catalog"
        self.repo_path = self.clone_catalog(output_path)
        self.pecha_catalog_file = self.repo_path / "opf_catalog.csv"
        self.catalog_data = read_csv(self.pecha_catalog_file)

    def clone_catalog(self, output_path: Path):
        repo_path = clone_repo("catalog", output_path)
        return repo_path

    def add_entry_to_pecha_catalog(self, new_entry: List[str]) -> None:
        """
        Update a Pecha information to PechaData pecha catalog
        """
        # Check if new entry already exists in the catalog data
        formated_new_entry = [
            str(data) if data is not None else "" for data in new_entry
        ]
        if formated_new_entry not in self.catalog_data:
            self.catalog_data.append(new_entry)
            write_csv(self.pecha_catalog_file, self.catalog_data)
            commit_and_push(
                Repo(self.repo_path), message="Update catalog", branch="main"
            )

    def get_pecha_id_with_title(self, title: str) -> str | None:
        """
        1.Get pecha/opf catalog data
        2.Compare title with each entry
        3.When matched, return the pecha id
        """
        for cat_data in self.catalog_data[1:]:
            pecha_id = cat_data[0]
            entry_title_str = cat_data[1]

            try:
                entry_title = ast.literal_eval(entry_title_str)
            except (ValueError, SyntaxError):
                entry_title = entry_title_str

            if isinstance(entry_title, dict):
                if title in entry_title.values():
                    return pecha_id

            elif isinstance(entry_title, str):
                if title == entry_title:
                    return pecha_id
        return None



from enum import Enum
from typing import Any, Dict, List, Optional

from openpecha.config import get_logger
from openpecha.pecha import Pecha
from openpecha.pecha.annotations import AnnBase
from openpecha.pecha.serializers.pecha_tools.translation_editor import (
    TranslationSerializer,
)

logger = get_logger(__name__)


class EditorType(Enum):
    """
    Editor Type for Serializer to determine the type of Pecha.
    """

    TRANSLATION = "translation"
    COMMENTARY = "commentary"
    CHAPTER = "chapter"
    SEGMENTATION = "segmentation"


class TranslationEditor:
    metadata: Dict = {}
    base_text: str = ""
    annotations: Optional[AnnBase] = None


class Serializer:
    def serialize(
        self,
        pechas: List[Pecha],
        metadatas: List[Dict[str, Any]],
        pecha_category: List[Dict],
        editor_type: str,
        layer_name: Optional[str] = None,
    ):
        """
        Serialize a Pecha based on its type.
        """
        logger.info(f"Serializing Pecha {pechas[0].id}, Editor Type: {editor_type}")

        match editor_type:
            case EditorType.TRANSLATION.value:
                return TranslationSerializer().serialize(
                    pechas, metadatas, pecha_category, layer_name
                )

            case _:
                raise ValueError(f"Unsupported pecha type: {editor_type}")



import re
from pathlib import Path
from typing import Dict, List, Tuple

from openpecha.config import PECHAS_PATH, get_logger
from openpecha.exceptions import FileNotFoundError, MetaDataValidationError
from openpecha.pecha import Pecha, annotation_path
from openpecha.pecha.layer import AnnotationType
from openpecha.pecha.metadata import InitialCreationType, PechaMetaData
from openpecha.pecha.parsers import DocxBaseParser
from openpecha.pecha.parsers.docx.utils import extract_text_from_docx

logger = get_logger(__name__)


class DocxRootParser(DocxBaseParser):
    def __init__(self):
        self.number_list_regex = r"^(\d+)\)\t(.*)"

    def extract_numbered_list(self, text: str) -> Dict[str, str]:
        """
        Extract number list from the extracted text from docx.

        Example Output:>
            {
                '1': 'དབུ་མ་དགོངས་པ་རབ་གསལ་ལེའུ་དྲུག་པ་བདེན་གཉིས་སོ་སོའི་ངོ་བོ་བཤད་པ།། ',
                '2': '2 གསུམ་པ་ལ་གཉིས། ཀུན་རྫོབ་ཀྱི་བདེན་པ་བཤད་པ་དང་། ',
                '3': '2,3 དེས་གང་ལ་སྒྲིབ་ན་ཡང་དག་ཀུན་རྫོབ་འདོད་ཅེས་པས་ཡང་དག་པའི་དོན་ལ་སྒྲིབ་པས་ཀུན་རྫོབ་བམ་སྒྲིབ་བྱེད་དུ་འདོད་ཅེས་པ་སྟེ། །',
                ...
            }
        """
        res: Dict[str, str] = {}
        for para_text in text.split("\n\n"):
            match = re.match(self.number_list_regex, para_text)
            if match:
                number = match.group(1)
                text = match.group(2)
                res[number] = text

        return res

    def calculate_segment_coordinates(
        self, segments: Dict[str, str]
    ) -> Tuple[List[Dict], str]:
        """Calculate start and end positions for each segment and build base text.

        Args:
            segments: Dictionary mapping root indices to segment text

        Returns:
            Tuple containing:
            - List of dicts with start/end positions for each segment
            - Combined base text with all segments
        """
        positions = []
        base = ""
        char_count = 0

        for root_idx_mapping, segment in segments.items():
            positions.append(
                {
                    "start": char_count,
                    "end": char_count + len(segment),
                    "root_idx_mapping": root_idx_mapping,
                }
            )
            base += f"{segment}\n"
            char_count += len(segment) + 1

        return (positions, base)

    def extract_segmentation_anns(
        self, positions: List[Dict[str, int]], ann_type: AnnotationType
    ) -> List[Dict]:
        """Create segment annotations from position information.

        Args:
            positions: List of dicts containing start/end positions and root index mappings

        Returns:
            List of annotation dictionaries
        """
        return [
            {
                ann_type.value: {
                    "start": pos["start"],
                    "end": pos["end"],
                },
                "root_idx_mapping": pos["root_idx_mapping"],
            }
            for pos in positions
        ]

    def extract_segmentation_coords(
        self, docx_file: Path
    ) -> Tuple[List[Dict[str, int]], str]:
        """Extract text from docx and calculate coordinates for segments.

        Args:
            docx_file: Path to the docx file

        Returns:
            Tuple containing:
            - List of dicts with segment positions and root index mappings
            - Base text containing all segments
        """
        # Extract and normalize text
        text = extract_text_from_docx(docx_file)
        logger.info(f"Extracted text: {text}")
        numbered_text = self.extract_numbered_list(text)
        logger.info(f"Extracted numbered list: {numbered_text}")
        return self.calculate_segment_coordinates(numbered_text)

    def parse(
        self,
        input: str | Path,
        annotation_type: AnnotationType,
        metadata: Dict,
        output_path: Path = PECHAS_PATH,
        pecha_id: str | None = None,
    ) -> Tuple[Pecha, annotation_path]:
        """Parse a docx file and create a pecha.

        The process is split into three main steps:
        1. Extract text and calculate coordinates
        2. Extract segmentation annotations
        3. Initialize pecha with annotations and metadata
        """
        input = Path(input)
        if not input.exists():
            logger.error(f"The input docx file {str(input)} does not exist.")
            raise FileNotFoundError(
                f"[Error] The input docx file '{str(input)}' does not exist."
            )

        output_path.mkdir(parents=True, exist_ok=True)

        positions, base = self.extract_segmentation_coords(input)

        pecha = self.create_pecha(base, output_path, metadata, pecha_id)
        annotation_path = self.add_segmentation_layer(pecha, positions, annotation_type)

        logger.info(f"Pecha {pecha.id} is created successfully.")
        return (pecha, annotation_path)

    def create_pecha(
        self, base: str, output_path: Path, metadata: Dict, pecha_id: str | None
    ) -> Pecha:
        pecha = Pecha.create(output_path, pecha_id)
        pecha.set_base(base)

        try:
            pecha_metadata = PechaMetaData(
                id=pecha.id,
                parser=self.name,
                **metadata,
                bases={},
                initial_creation_type=InitialCreationType.google_docx,
            )
        except Exception as e:
            logger.error(f"The metadata given was not valid. {str(e)}")
            raise MetaDataValidationError(
                f"[Error] The metadata given was not valid. {str(e)}"
            )
        else:
            pecha.set_metadata(pecha_metadata.to_dict())

        return pecha

    def add_segmentation_layer(
        self, pecha: Pecha, positions: List[Dict], ann_type: AnnotationType
    ) -> annotation_path:

        basename = list(pecha.bases.keys())[0]
        layer, layer_path = pecha.add_layer(basename, ann_type)
        anns = self.extract_segmentation_anns(positions, ann_type)
        for ann in anns:
            pecha.add_annotation(layer, ann, ann_type)
        layer.save()

        return str(layer_path.relative_to(pecha.layer_path))



import re
from pathlib import Path
from typing import Any, Dict, List, Optional

from docx import Document
from docx.shared import RGBColor

from openpecha.config import PECHAS_PATH
from openpecha.pecha import Pecha
from openpecha.pecha.layer import AnnotationType
from openpecha.pecha.metadata import InitialCreationType
from openpecha.pecha.parsers import BaseParser
from openpecha.pecha.parsers.parser_utils import extract_metadata_from_xlsx


class DocxComplexCommentaryParser(BaseParser):
    def __init__(self, root_path: Optional[str] = None):
        self.root_path = root_path
        self.commentary_segment_splitter = "\n\n"
        self.meaning_segment_anns: List[Dict[str, Any]] = []
        self.sapche_anns: List[Dict[str, Any]] = []
        self.temp_state = {
            "meaning_segment": {"anns": [], "char_diff": 0},
            "sapche": {"anns": [], "char_diff": 0},
        }
        self.base = ""
        self.metadata: Dict[str, Any] = {}

    def normalize_text(self, text: str):
        text = self.normalize_whitespaces(text)
        text = self.normalize_newlines(text)
        return text

    @staticmethod
    def normalize_whitespaces(text: str):
        """
        If there are spaces or tab between newlines, it will be removed.
        """
        return re.sub(r"\n[\s\t]+\n", "\n\n", text)

    @staticmethod
    def normalize_newlines(text: str):
        """
        If there are more than 2 newlines continuously, it will replace it with 2 newlines.
        """
        return re.sub(r"\n{3,}", "\n\n", text)

    def parse(
        self,
        input: Path,
        metadata: Dict[str, Any] | Path,
        output_path: Path = PECHAS_PATH,
    ) -> Pecha:

        # Clean up class attributes
        self.meaning_segment_anns = []
        self.base = ""

        if isinstance(metadata, Path):
            metadata = extract_metadata_from_xlsx(metadata)

        assert isinstance(metadata, dict)
        metadata["root_path"] = self.root_path
        self.metadata = metadata
        self.parse_commentary(input)

        pecha = self.create_pecha(output_path)
        return pecha

    def prepare_doc(self, input: Path):
        """
        Input: a docx file
        Process: Prepare the doc for parsing
        """

        def format_paragraphs(paragraphs: List[Dict[str, Any]]) -> Dict[str, Any]:
            """
            Paragraphs is a text with styles.
            Each line in docx file is a paragraph.
            We have to combine the text and the styles
            """
            formatted_paras = []
            para_texts = []
            for para in paragraphs:
                para_texts.append(para["text"].strip())
                style_texts = []
                styles = []
                for para_style in para["styles"]:
                    para_text = para_style.text
                    style_texts.append(para_text)
                    styles.append(para_style.font)
                if style_texts:
                    style_texts[0] = style_texts[0].lstrip()
                    style_texts[-1] = style_texts[-1].rstrip()
                formatted_paras.append({"texts": style_texts, "styles": styles})
            res = {"text": "\n".join(para_texts), "styles": formatted_paras}
            return res

        # Parse the document
        docs = Document(input)

        formatted_docs = []
        last_doc_data: List[Dict[str, Any]] = []

        for doc in docs.paragraphs:
            if doc.text.strip() == "":
                if last_doc_data:
                    formatted_docs.append(format_paragraphs(last_doc_data))
                    last_doc_data = []
            else:
                last_doc_data.append({"text": doc.text, "styles": doc.runs})

        # Handle remaining paragraphs after the loop
        if last_doc_data:
            formatted_docs.append(format_paragraphs(last_doc_data))

        return formatted_docs

    @staticmethod
    def update_doc(doc: Dict[str, Any], char_diff: int):
        """
        Updates the document by removing characters up to the given char_diff.
        Args:
            doc (Dict[str, Any]): The document to update, containing text and styles.
            char_diff (int): The number of characters to remove from the beginning of the text.
        Returns:
            Dict[str, Any]: The updated document.
        """
        # Update the main text field
        doc["text"] = doc["text"][char_diff:]

        # Extract the first style's texts and styles
        styles = doc["styles"][0]
        texts = styles["texts"]
        style_meta = styles["styles"]

        char_count = 0
        for idx, text_chunk in enumerate(texts):
            if char_count >= char_diff or char_count + len(text_chunk) == char_diff:
                doc["styles"][0]["styles"] = style_meta[idx + 1 :]
                doc["styles"][0]["texts"] = texts[idx + 1 :]
                break

            if char_count + len(text_chunk) > char_diff:
                doc["styles"][0]["styles"] = style_meta[idx:]
                doc["styles"][0]["texts"] = [
                    text_chunk[char_diff - char_count :]
                ] + texts[idx + 1 :]
                break
            char_count += len(text_chunk)

        return doc

    def add_commentary_meaning_ann(self, doc: Dict[str, Any], char_count: int):
        segment = doc["text"]
        match = re.match(r"^([\d\-,]+) ", segment)
        updated_segment = segment
        if match:
            root_idx_mapping = match.group(1)
            segment = segment.replace(root_idx_mapping, "")
            doc = self.update_doc(doc, len(root_idx_mapping) + 1)
            updated_segment = segment.strip()
            curr_segment_ann = {
                AnnotationType.ALIGNMENT.value: {
                    "start": char_count,
                    "end": char_count + len(updated_segment),
                },
                "root_idx_mapping": root_idx_mapping,
            }
        else:
            curr_segment_ann = {
                AnnotationType.ALIGNMENT.value: {
                    "start": char_count,
                    "end": char_count + len(segment),
                }
            }

        self.temp_state["meaning_segment"]["anns"].append(curr_segment_ann)  # type: ignore
        return doc

    def add_sapche_ann(self, doc: Dict[str, Any], char_count: int):
        """
        Extract and process sapche annotations (in Fuchsia/Pink color) from the provided docx file structure.
        Args:
            doc (Dict[str, Any]): The document structure containing styles and text.
            char_count (int): The initial character count for span calculation.
        Returns:
            str: The updated segment text after processing annotations.
        """
        inner_char_count = 0
        sapche_anns: List[Dict[str, Any]] = []
        for doc_style in doc["styles"]:
            for idx in range(len(doc_style["texts"])):
                if doc_style["styles"][idx].color.rgb == RGBColor(0xFF, 0x00, 0x00):
                    match = re.match(r"([\d\.]+)\s", doc_style["texts"][idx])
                    if match:
                        # Extract sapche number and store the char length to update the previous ann spans
                        sapche_number = match.group(1)
                        doc_style["texts"][idx] = doc_style["texts"][idx].replace(
                            f"{sapche_number} ", ""
                        )
                        self.temp_state["sapche"]["char_diff"] += len(sapche_number)  # type: ignore

                        start = char_count + inner_char_count
                        end = start + len(doc_style["texts"][idx])
                        sapche_anns.append(
                            {
                                AnnotationType.SAPCHE.value: {
                                    "start": start,
                                    "end": end,
                                },
                                "sapche_number": sapche_number,
                            }
                        )
                    # If the sapche number is not needed, use the following code in future
                    # else:
                    #     start = char_count + inner_char_count
                    #     end = start + len(doc_style["texts"][idx])
                    #     sapche_anns.append(
                    #         {
                    #             AnnotationType.sapche.value: {
                    #                 "start": start,
                    #                 "end": end,
                    #             }
                    #         }
                    #     )
                inner_char_count += len(doc_style["texts"][idx])
            inner_char_count += 1  # for newline

        formatted_anns = self.merge_anns(sapche_anns, AnnotationType.SAPCHE)
        self.temp_state["sapche"]["anns"].extend(formatted_anns)  # type: ignore
        updated_segment = "\n".join(
            ["".join(doc_style["texts"]) for doc_style in doc["styles"]]
        )
        return updated_segment

    @staticmethod
    def merge_anns(
        anns: List[Dict[str, Any]], ann_layer: AnnotationType
    ) -> List[Dict[str, Any]]:
        """
        Merge overlapping or consecutive sapche annotations.
        Args:
            annotations (List[Dict[str, Any]]): Eg: List of sapche annotations.

        Returns:
            List[Dict[str, Any]]: Merged annotations.
        """
        formatted_anns: List[Dict[str, Any]] = []
        last_ann: Optional[Dict[str, Any]] = None
        for ann in anns:
            if last_ann is None:
                last_ann = ann
                continue
            if ann[ann_layer.value]["start"] != last_ann[ann_layer.value]["end"]:
                formatted_anns.append(last_ann)
                last_ann = ann
            else:
                last_ann[ann_layer.value]["end"] = ann[ann_layer.value]["end"]

        if last_ann:
            formatted_anns.append(last_ann)
        return formatted_anns

    def update_ann_spans(self):
        """
        Update the spans of the meaning_segment and sapche annotations.
        """
        if self.temp_state["meaning_segment"]["anns"]:
            meaning_segment_ann = self.temp_state["meaning_segment"]["anns"][0]  # type: ignore
            meaning_segment_ann[AnnotationType.ALIGNMENT.value][
                "end"
            ] -= self.temp_state["sapche"]["char_diff"]
            self.meaning_segment_anns.append(meaning_segment_ann)

        self.sapche_anns.extend(self.temp_state["sapche"]["anns"])  # type: ignore

        self.temp_state = {
            "meaning_segment": {"anns": [], "char_diff": 0},
            "sapche": {"anns": [], "char_diff": 0},
        }

    def parse_commentary(self, input: Path):
        """
        Input: a docx file
        Process: - Parse and record the commentary annotations in self.meaning_segment_anns,
                 - Save the cleaned base text in self.base
        """
        formatted_docs = self.prepare_doc(input)

        char_count = 0
        base_texts = []
        for doc in formatted_docs:
            segment = doc["text"]
            if not segment:
                continue

            doc = self.add_commentary_meaning_ann(doc, char_count)
            updated_segment = self.add_sapche_ann(doc, char_count)

            self.update_ann_spans()

            base_texts.append(updated_segment)
            char_count += len(updated_segment)
            char_count += 2  # for two newlines

        self.base = "\n\n".join(base_texts)

    def create_pecha(self, output_path: Path) -> Pecha:
        pecha = Pecha.create(output_path)
        basename = pecha.set_base(self.base)

        # Add meaning_segment layer
        meaning_segment_layer, _ = pecha.add_layer(basename, AnnotationType.ALIGNMENT)
        for ann in self.meaning_segment_anns:
            pecha.add_annotation(meaning_segment_layer, ann, AnnotationType.ALIGNMENT)
        meaning_segment_layer.save()

        # Add sapche layer
        sapche_layer, _ = pecha.add_layer(basename, AnnotationType.SAPCHE)
        for ann in self.sapche_anns:
            pecha.add_annotation(sapche_layer, ann, AnnotationType.SAPCHE)
        sapche_layer.save()

        pecha.set_metadata(
            {
                "id": pecha.id,
                "parser": self.name,
                "initial_creation_type": InitialCreationType.google_docx,
                **self.metadata,
            }
        )

        return pecha



from datetime import datetime
from enum import Enum
from typing import Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field, field_serializer, model_validator

from openpecha.ids import get_diplomatic_id, get_initial_pecha_id, get_open_pecha_id


class InitialCreationType(Enum):
    ocr = "ocr"
    ebook = "ebook"
    input = "input"
    tmx = "tmx"
    json = "json"
    google_docx = "google_docx"


class Language(Enum):
    tibetan = "bo"
    english = "en"
    literal_chinese = "lzh"
    chinese = "zh"
    sanskrit = "sa"
    italian = "it"
    russian = "ru"
    hindi = "hi"


class CopyrightStatus(Enum):
    UNKNOWN = "Unknown"
    COPYRIGHTED = "In copyright"
    PUBLIC_DOMAIN = "Public domain"


class Copyright(BaseModel):
    status: CopyrightStatus = CopyrightStatus.UNKNOWN  # noqa
    notice: Optional[str] = ""
    info_url: Optional[str] = None

    model_config = ConfigDict(extra="forbid")


Copyright_copyrighted = Copyright(
    status=CopyrightStatus.COPYRIGHTED,
    notice="In copyright by the original author or editor",
    info_url="http://rightsstatements.org/vocab/InC/1.0/",
)

Copyright_unknown = Copyright(
    status=CopyrightStatus.UNKNOWN,
    notice="Copyright Undertermined",
    info_url="http://rightsstatements.org/vocab/UND/1.0/",
)

Copyright_public_domain = Copyright(
    status=CopyrightStatus.PUBLIC_DOMAIN,
    notice="Public domain",
    info_url="https://creativecommons.org/publicdomain/mark/1.0/",
)


class LicenseType(Enum):
    # based on https://creativecommons.org/licenses/

    CC0 = "CC0"
    PUBLIC_DOMAIN_MARK = "Public Domain Mark"
    CC_BY = "CC BY"
    CC_BY_SA = "CC BY-SA"
    CC_BY_ND = "CC BY-ND"
    CC_BY_NC = "CC BY-NC"
    CC_BY_NC_SA = "CC BY-NC-SA"
    CC_BY_NC_ND = "CC BY-NC-ND"

    UNDER_COPYRIGHT = "under copyright"
    UNKNOWN = "Unknown"


class PechaMetaData(BaseModel):
    id: str
    title: Optional[Dict[str, str] | str] = None
    author: Optional[List[str] | Dict[str, str] | str] = None
    imported: Optional[datetime] = None
    source: Optional[str] = None
    toolkit_version: str
    parser: str
    initial_creation_type: InitialCreationType
    language: Optional[Language] = None
    source_metadata: Dict = {}
    bases: Dict = {}
    copyright: Copyright = Copyright()
    licence: LicenseType = LicenseType.UNKNOWN

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
    # Optional fields from both classes
    legacy_id: Optional[str] = None
    source_file: Optional[str] = None

    # Metadata fields
    ocr_import_info: Optional[Dict] = None
    statistics: Optional[Dict] = None
    quality: Optional[Dict] = None

    # Time tracking
    last_modified: Optional[datetime] = None

    # Validators from both classes
    @model_validator(mode="before")
    def set_id(cls, values):
        if "id" not in values or values["id"] is None:
            values["id"] = get_initial_pecha_id()
        return values

    # @model_validator(mode="before")
    # def validate_parser(cls, values):
    #     if "parser" in values and values["parser"]:
    #         parser_classes = cls.get_toolkit_parsers()
    #         if values["parser"] not in [name for name, _ in parser_classes]:
    #             raise ValueError(
    #                 f"Parser {values['parser']} not in the Toolkit parsers."
    #             )
    #     return values

    @model_validator(mode="before")
    def set_toolkit_version(cls, values):
        if "toolkit_version" not in values or values["toolkit_version"] is None:
            try:
                from importlib.metadata import PackageNotFoundError, version

                toolkit_version = version("openpecha")
                values["toolkit_version"] = toolkit_version
            except PackageNotFoundError as e:
                raise RuntimeError("Package 'openpecha' not found.") from e
            except Exception as e:
                raise RuntimeError(f"Error fetching toolkit version: {str(e)}") from e
        return values

    @model_validator(mode="before")
    def set_imported(cls, values):
        if "imported" not in values or values["imported"] is None:
            values["imported"] = datetime.now()
        return values

    @model_validator(mode="before")
    def set_last_modified(cls, values):
        if "last_modified" not in values or values["last_modified"] is None:
            values["last_modified"] = datetime.now()
        return values

    @model_validator(mode="before")
    def set_copyright_info(cls, values):
        if "copyright" not in values or values["copyright"] is None:
            values["copyright"] = Copyright()
        return values

    # Serializers for complex types
    @field_serializer("imported", mode="plain")
    def serialize_imported(self, value: Optional[datetime]) -> Optional[str]:
        return value.isoformat() if value else None

    @field_serializer("last_modified", mode="plain")
    def serialize_last_modified(self, value: Optional[datetime]) -> Optional[str]:
        return value.isoformat() if value else None

    @field_serializer("licence", mode="plain")
    def serialize_licence(self, value: Optional[LicenseType]) -> Optional[str]:
        return value.value if value else None

    @field_serializer("language", mode="plain")
    def serialize_language(self, value: Optional[Language]) -> Optional[str]:
        return value.value if value else None

    @field_serializer("initial_creation_type", mode="plain")
    def serialize_initial_creation_type(
        self, value: Optional[InitialCreationType]
    ) -> Optional[str]:
        return value.value if value else None

    @field_serializer("copyright", mode="plain")
    def serialize_copyright(self, value: Optional[Copyright]) -> Optional[Dict]:
        if not value:
            return None
        return {
            "status": value.status.value,
            "notice": value.notice,
            "info_url": value.info_url,
        }

    def update_last_modified_date(self):
        self.last_modified = datetime.now()

    # @classmethod
    # def get_toolkit_parsers(cls):
    #     # List to store all classes from the package
    #     all_classes = []
    #     import sys

    #     base_path = Path(__file__).parent / "parsers"
    #     pecha_parser_path = "openpecha.pecha.parsers"

    #     for py_file in base_path.rglob("*.py"):
    #         path_parts = list(py_file.parts)
    #         path_parts[-1] = path_parts[-1].replace(".py", "")
    #         if path_parts[-1] == "__init__":
    #             path_parts.pop()

    #         if path_parts[0] == "/":
    #             path_parts.pop(0)

    #         start_index = path_parts.index(pecha_parser_path.split(".")[0])
    #         parser_path = ".".join(path_parts[start_index:])
    #         importlib.import_module(parser_path)
    #         classes = inspect.getmembers(sys.modules[parser_path], inspect.isclass)
    #         all_classes.extend(classes)

    #     parsers = importlib.import_module("openpecha.pecha.parsers")
    #     parser_classes = [
    #         (name, class_)
    #         for name, class_ in all_classes
    #         if issubclass(class_, parsers.BaseParser)
    #         or issubclass(class_, parsers.OCRBaseParser)
    #         and class_ is not parsers.BaseParser
    #         and class_ is not parsers.OCRBaseParser
    #     ]
    #     return parser_classes

    def to_dict(self):
        """
        Prepare PechaMetaData attribute to be JSON serializable
        """
        data = self.model_dump()

        # Dynamically get standard fields from the model
        standard_fields = list(set(type(self).model_fields.keys()))

        # Move any extra fields to source_metadata
        extra_fields = {}
        for k, v in data.items():
            if k not in standard_fields:
                if isinstance(v, Enum):
                    extra_fields[k] = v.value
                elif isinstance(v, datetime):
                    extra_fields[k] = v.isoformat()
                else:
                    extra_fields[k] = v

        if "source_metadata" not in data:
            data["source_metadata"] = {}
        data["source_metadata"].update(extra_fields)

        # Remove extra fields from the top-level data
        for field in extra_fields:
            del data[field]

        return data


class InitialPechaMetadata(PechaMetaData):
    @model_validator(mode="before")
    def set_id(cls, values):
        if "id" not in values or values["id"] is None:
            values["id"] = get_initial_pecha_id()
        return values


class OpenPechaMetadata(PechaMetaData):
    @model_validator(mode="before")
    def set_id(cls, values):
        if "id" not in values or values["id"] is None:
            values["id"] = get_open_pecha_id()
        return values


class DiplomaticPechaMetadata(PechaMetaData):
    @model_validator(mode="before")
    def set_id(cls, values):
        if "id" not in values or values["id"] is None:
            values["id"] = get_diplomatic_id()
        return values


class KungsangMonlamMetaData(BaseModel):
    author: Optional[Dict[str, str]] = Field(default_factory=dict)
    composition_date: Optional[Dict[str, str]] = Field(default_factory=dict)
    source: Optional[Dict[str, str]] = Field(default_factory=dict)
    presentation: Optional[Dict[str, str]] = Field(default_factory=dict)
    usage_title: Optional[Dict[str, str]] = Field(default_factory=dict)
    title_short: Optional[Dict[str, str]] = Field(default_factory=dict)
    title_long_clean: Optional[Dict[str, str]] = Field(default_factory=dict)
    title_alt_1: Optional[Dict[str, str]] = Field(default_factory=dict)
    title_alt_2: Optional[Dict[str, str]] = Field(default_factory=dict)
    is_commentary_of: Optional[Dict[str, str]] = Field(default_factory=dict)
    translation_of: Optional[Dict[str, str]] = Field(default_factory=dict)
    lang: Optional[str] = None

    model_config = ConfigDict(extra="allow")

    def to_pecha_metadata(self, parser: str) -> PechaMetaData:
        """
        Extract relevant fields from KunsangMonlamMetaData and map them to PechaMetaData fields
        """
        title = self.title_short if self.title_short else {}
        author = self.author if self.author else []
        source = self.source if self.source else ""
        language = self.lang if self.lang else ""

        extra_metadata = {
            k: v
            for k, v in self.model_dump().items()
            if k not in ["author", "source", "source", "lang"]
        }

        return PechaMetaData(
            title=title,
            author=author,
            source=source,
            parser=parser,
            initial_creation_type=InitialCreationType.input,
            language=language,
            source_metadata={},
            **extra_metadata,
        )



import re
from pathlib import Path
from typing import Dict, List

from botok.tokenizers.chunktokenizer import ChunkTokenizer

from openpecha.config import PECHAS_PATH
from openpecha.pecha import Pecha
from openpecha.pecha.layer import AnnotationType
from openpecha.pecha.parsers import BaseParser
from openpecha.utils import read_json


class PedurmaParser(BaseParser):
    def __init__(self):
        self.ann_regex = r"(\(\d+\) <.+?>)"
        self.pagination_regex = r"\d+-\d+"
        self.base_text = ""
        self.pedurma_anns = []
        self.meaning_segment_anns = []

    def get_base_text(self, text: str):
        text = re.sub(self.ann_regex, "", text)
        text = text.replace(":", "")
        return text

    def parse(
        self,
        input: str,
        metadata: Dict | Path,
        output_path: Path = PECHAS_PATH,
    ) -> Pecha:

        # Remove pagination
        input = re.sub(self.pagination_regex, "", input)
        self.base_text = self.get_base_text(input)

        # Normalize newlines with #
        input = input.replace("\n", "#")
        char_walker = 0
        # Split the text into chunks with anns regex
        chunks = re.split(self.ann_regex, input)
        prev_chunk = chunks[0]
        self.pedurma_anns = []
        for chunk in chunks:
            if re.search(self.ann_regex, chunk):
                ann = get_annotation(prev_chunk, chunk, char_walker)
                self.pedurma_anns.append(ann)
            else:
                clean_chunk = chunk.replace(":", "")
                char_walker += len(clean_chunk)
            prev_chunk = chunk

        input = input.replace("#", "\n")

        # Segment Annotation
        char_walker = 0
        self.meaning_segment_anns = []
        for line in self.base_text.splitlines():
            segment_ann = {
                AnnotationType.SEGMENTATION.value: {
                    "start": char_walker,
                    "end": char_walker + len(line),
                }
            }
            self.meaning_segment_anns.append(segment_ann)
            char_walker += len(line)
            char_walker += 1  # Add because of new line

        # Create a pecha
        pecha = Pecha.create(output_path)

        basename = pecha.set_base(self.base_text)

        # Add Durchen Layer
        durchen_layer, _ = pecha.add_layer(basename, AnnotationType.DURCHEN)
        for ann in self.pedurma_anns:
            pecha.add_annotation(durchen_layer, ann, AnnotationType.DURCHEN)

        durchen_layer.save()

        # Add Segment Layer
        segment_layer, _ = pecha.add_layer(basename, AnnotationType.SEGMENTATION)
        for ann in self.meaning_segment_anns:
            pecha.add_annotation(segment_layer, ann, AnnotationType.SEGMENTATION)

        segment_layer.save()
        # Set metadata
        if isinstance(metadata, Path):
            metadata = read_json(metadata)

        assert isinstance(metadata, dict)
        metadata = modify_metadata(metadata)
        pecha.set_metadata({"id": pecha.id, "parser": self.name, **metadata})  # noqa

        return pecha


def modify_metadata(metadata: Dict) -> Dict:
    modified_metadata = {
        k: v
        for k, v in metadata.items()
        if k not in ["title_bo", "alt_title_bo", "author_en", "author_bo"]
    }
    modified_metadata["title"] = {
        "title_bo": metadata["title_bo"],
        "alt_title_bo": metadata["alt_title_bo"],
    }
    modified_metadata["author"] = {
        "author_en": metadata["author_en"],
        "author_bo": metadata["author_bo"],
    }
    return modified_metadata


def get_annotation(prev_chunk: str, note_chunk: str, char_walker: int):
    span_text = get_span_text(prev_chunk, note_chunk)
    start = char_walker - len(span_text)
    end = char_walker

    ann = {
        AnnotationType.DURCHEN.value: {"start": start, "end": end},
        "note": note_chunk,
    }
    return ann


def get_span_text(prev_chunk: str, note_chunk: str):
    """
    Input: text chunk
    Process: Extract span text where the syllable/words have variations
             if ':' is present, extract the text after ':'
             else extract the last syllable
    Output: span text

    Example:
    Input: །དེ་བཞིན་ཉོན་མོངས་རྣམ་  Output: རྣམ་
    Input: ཆོས་ཀྱི་དབྱིངས་སུ་བསྟོད་པ། :འཕགས་པ་འཇམ་ Output: འཕགས་པ་འཇམ་
    """
    span_text = ""
    if "+" in note_chunk:
        return span_text
    if ":" in prev_chunk:
        match = re.search(":.*", prev_chunk)
        if match:
            span_text = match.group(
                0
            )  # Use group(0) to safely access the matched string
    else:
        syls = get_syls(prev_chunk)
        if syls:
            span_text = syls[-1]
            if span_text == "#":
                span_text = syls[-2]
    span_text = span_text.replace("#", "\n")
    span_text = span_text.replace(":", "")
    return span_text


def get_syls(text: str):
    """
    Split the text into syllables
    """
    tokenizer = ChunkTokenizer(text)

    tokens = tokenizer.tokenize()
    syls: List = []
    syl_walker = 0
    for token in tokens:
        token_string = token[1]
        if is_shad(token_string):
            try:
                syls[syl_walker - 1] += token_string
            except:  # noqa
                syls.append(token_string)
                syl_walker += 1
        else:
            syls.append(token_string)
            syl_walker += 1
    return syls


def is_shad(text):
    shads = ["། །", "།", "།།", "། "]
    if text in shads:
        return True
    return False


"""
PEDURMA PREPROCESSING
"""


def filter_pedurma_ann(text: str):
    # remove <«སྣར་»«པེ་»བསྒྲུབ་པའི་> from text
    text = re.sub(r"<[\u0F00-\u0FFF\s«»]+>", "", text)
    # remove numbering (28) from text
    text = re.sub(r"\(\d+\)", "", text)

    return text


def split_by_shad(text: str):
    return text.split("།")


def preprocess_pedurma_text(text: str):
    from bo_sent_tokenizer import segment
    from fast_antx.core import transfer

    filtered_text = filter_pedurma_ann(text)
    tokenized_text = segment(filtered_text, keep_non_bo_and_symbols=True)

    annotations = [["note_transfer", r"(\(\d+\) <[\u0F00-\u0FFF\s«»]+>)"]]
    preprocessed_text = transfer(text, annotations, tokenized_text, output="txt")

    return preprocessed_text



from openpecha.utils import (
    adjust_segment_num_for_chapter,
    chunk_strings,
    get_chapter_for_segment,
    parse_root_mapping,
)


def test_parse_root_mapping():
    input = "1"
    assert parse_root_mapping(input) == [1]

    input = "1,2,3,4"
    assert parse_root_mapping(input) == [1, 2, 3, 4]

    input = "1-4"
    assert parse_root_mapping(input) == [1, 2, 3, 4]

    input = "1-4,5-8"
    assert parse_root_mapping(input) == [1, 2, 3, 4, 5, 6, 7, 8]


def test_chunk_strings():
    # Less than chunk_size
    strings = ["1", "2", "3", "4", "5"]
    chunk_size = 10
    expected = [["1", "2", "3", "4", "5"]]
    assert chunk_strings(strings, chunk_size) == expected

    # More than chunk_size
    strings = ["1", "2", "3", "4", "5", "6", "7", "8", "9", "10"]
    chunk_size = 3
    expected = [["1", "2", "3"], ["4", "5", "6"], ["7", "8", "9"], ["10"]]
    assert chunk_strings(strings, chunk_size) == expected

    # Equal to chunk_size
    strings = ["1", "2", "3", "4", "5"]
    chunk_size = 5
    expected = [["1", "2", "3", "4", "5"]]
    assert chunk_strings(strings, chunk_size) == expected

    # Empty list
    strings = []
    chunk_size = 5
    expected = []
    assert chunk_strings(strings, chunk_size) == expected

    # Evenly divisible
    strings = ["1", "2", "3", "4", "5", "6", "7", "8", "9", "10"]
    chunk_size = 2
    expected = [["1", "2"], ["3", "4"], ["5", "6"], ["7", "8"], ["9", "10"]]
    assert chunk_strings(strings, chunk_size) == expected


def test_get_chapter_num_from_segment_num():
    segment_num = 1
    no_of_chapter_segment = 100
    assert get_chapter_for_segment(segment_num, no_of_chapter_segment) == 1

    segment_num = 100
    no_of_chapter_segment = 100
    assert get_chapter_for_segment(segment_num, no_of_chapter_segment) == 1

    segment_num = 101
    no_of_chapter_segment = 100
    assert get_chapter_for_segment(segment_num, no_of_chapter_segment) == 2

    segment_num = 200
    no_of_chapter_segment = 100
    assert get_chapter_for_segment(segment_num, no_of_chapter_segment) == 2

    segment_num = 893
    no_of_chapter_segment = 100
    assert get_chapter_for_segment(segment_num, no_of_chapter_segment) == 9


def test_process_segment_num_for_chapter():
    segment_num = 1
    no_of_chapter_segment = 100
    assert adjust_segment_num_for_chapter(segment_num, no_of_chapter_segment) == 1

    segment_num = 100
    no_of_chapter_segment = 100
    assert adjust_segment_num_for_chapter(segment_num, no_of_chapter_segment) == 100

    segment_num = 101
    no_of_chapter_segment = 100
    assert adjust_segment_num_for_chapter(segment_num, no_of_chapter_segment) == 1

    segment_num = 200
    no_of_chapter_segment = 100
    assert adjust_segment_num_for_chapter(segment_num, no_of_chapter_segment) == 100

    segment_num = 893
    no_of_chapter_segment = 100
    assert adjust_segment_num_for_chapter(segment_num, no_of_chapter_segment) == 93



import re

from openpecha.ids import (
    get_alignment_id,
    get_annotation_id,
    get_base_id,
    get_collection_id,
    get_diplomatic_id,
    get_id,
    get_initial_pecha_id,
    get_layer_id,
    get_open_pecha_id,
    get_uuid,
    get_work_id,
)


def test_get_uuid():
    uuid = get_uuid()
    assert re.match(
        r"^[0-9a-fA-F]{32}$", uuid
    ), f"UUID {uuid} is not in the correct format"


def test_get_id():
    prefix = "T"
    length = 4
    generated_id = get_id(prefix, length)
    assert re.match(
        r"^T[0-9A-F]{4}$", generated_id
    ), f"ID {generated_id} is not in the correct format"


def test_get_base_id():
    base_id = get_base_id()
    assert re.match(
        r"^[0-9A-F]{4}$", base_id
    ), f"Base ID {base_id} is not in the correct format"


def test_get_layer_id():
    layer_id = get_layer_id()
    assert re.match(
        r"^[0-9A-F]{4}$", layer_id
    ), f"Layer ID {layer_id} is not in the correct format"


def test_get_initial_pecha_id():
    initial_pecha_id = get_initial_pecha_id()
    assert re.match(
        r"^I[0-9A-F]{8}$", initial_pecha_id
    ), f"Initial Pecha ID {initial_pecha_id} is not in the correct format"


def test_get_open_pecha_id():
    open_pecha_id = get_open_pecha_id()
    assert re.match(
        r"^O[0-9A-F]{8}$", open_pecha_id
    ), f"Open Pecha ID {open_pecha_id} is not in the correct format"


def test_get_diplomatic_id():
    diplomatic_id = get_diplomatic_id()
    assert re.match(
        r"^D[0-9A-F]{8}$", diplomatic_id
    ), f"Diplomatic ID {diplomatic_id} is not in the correct format"


def test_get_work_id():
    work_id = get_work_id()
    assert re.match(
        r"^W[0-9A-F]{8}$", work_id
    ), f"Work ID {work_id} is not in the correct format"


def test_get_alignment_id():
    alignment_id = get_alignment_id()
    assert re.match(
        r"^A[0-9A-F]{8}$", alignment_id
    ), f"Alignment ID {alignment_id} is not in the correct format"


def test_get_collection_id():
    collection_id = get_collection_id()
    assert re.match(
        r"^C[0-9A-F]{8}$", collection_id
    ), f"Collection ID {collection_id} is not in the correct format"


def test_get_annotation_id():
    ann_id = get_annotation_id()
    assert len(ann_id) == 10



from bs4 import BeautifulSoup

from openpecha.config import get_logger
from openpecha.pecha.parsers.ocr.ocr import BBox, OCRParser

# Initialize the logger
logger = get_logger(__name__)


class HOCRParser(OCRParser):
    """
    OpenPecha Formatter for Google OCR HOCR output of scanned pecha.
    """

    def __init__(self, mode=None, output_path=None, metadata=None):
        super().__init__(output_path, metadata)
        self.mode = mode
        self.word_span = 0

    def get_word_text_with_space(self, line_text, word_box):
        """check for space after word_text using line_text, add space to word_text if found"""
        text = word_box.text
        self.word_span += len(text)
        if len(line_text) == self.word_span:
            return text
        else:
            next_character = line_text[self.word_span]
            if next_character == " ":
                self.word_span += 1
                text = text + " "
        return text

    def parse_box(self, line_box, word_box):
        """parse the word_box to create bbox for the text in word_box.

        Args:
             line_text (_type_): line from the page html
             word_box (_type_): word from the above line

         Returns:
             box : bbox for text in word_box with vertices, confidence, language
        """
        line_text = line_box.text
        if not word_box.has_attr("title"):
            return None
        boxinfos = word_box["title"].split(";")
        coords = None
        angle = None
        confidence = None
        for boxinfo in boxinfos:
            boxinfo_parts = boxinfo.strip().split(" ")
            if boxinfo_parts[0] == "bbox":
                # in HOCR's, bbox order is x0, y0, x1, y1
                coords = [
                    int(boxinfo_parts[1]),
                    int(boxinfo_parts[2]),
                    int(boxinfo_parts[3]),
                    int(boxinfo_parts[4]),
                ]
            elif boxinfo_parts[0] == "textangle":
                # angle is indicated counter-clockwise in hocr so
                # we need to convert it to our internal value system:
                angle = int(boxinfo_parts[1])
                if angle != 0:
                    angle = 360 - angle
            elif boxinfo_parts[0] == "x_wconf":
                confidence = float(boxinfo_parts[1]) / 100.0
        if coords is None:
            return None
        if self.remove_rotated_boxes and angle is not None and angle > 0:
            return None
        language = self.get_main_language_code(word_box.text)
        text = self.get_word_text_with_space(line_text, word_box)
        # but we initialize as x1, x2, y1, y2
        box = BBox(
            coords[0],
            coords[2],
            coords[1],
            coords[3],
            angle=angle,
            text=text,
            confidence=confidence,
            language=language,
        )
        return box

    def get_boxes(self, hocr_page_html):
        """parse the hocr_page_html for one html file per page for bboxes of the page

        Args:
            hocr_page_html (html): html of the page

        Returns:
            boxes: list of bbox dict
        """
        bboxes = []
        hocr_html = BeautifulSoup(hocr_page_html, "html.parser")
        line_boxes = hocr_html.find_all("span", {"class": "ocr_line"})
        for line_box in line_boxes:
            cur_line_boxes = []
            self.word_span = 0
            word_boxes = line_box.find_all("span", {"class": "ocrx_word"})
            for word_box in word_boxes:
                bbox = self.parse_box(line_box, word_box)
                if bbox is not None:
                    cur_line_boxes.append(bbox)
            bboxes.append(cur_line_boxes)
        return bboxes

    def get_boxes_for_IA(self, page_html):
        """parse the page_html for one html file per volume for bboxes of the page

        Args:
            hocr_page_html (html): html of the page

        Returns:
            boxes: list of bbox dict
        """
        bboxes = []
        paragraphs_html = page_html.find_all("p", {"class": "ocr_par"})
        for paragraph_html in paragraphs_html:
            line_boxes = paragraph_html.find_all("span", {"class": "ocr_line"})
            for line_box in line_boxes:
                cur_line_boxes = []
                self.word_span = 0
                word_boxes = line_box.find_all("span", {"class": "ocrx_word"})
                for word_box in word_boxes:
                    bbox = self.parse_box(line_box, word_box)
                    if bbox is not None:
                        cur_line_boxes.append(bbox)
                bboxes.append(cur_line_boxes)
        return bboxes

    def get_bboxes_for_page(self, image_group_id, image_filename):
        bboxes = []
        hocr_page_html = self.data_source.get_image_data(image_group_id, image_filename)
        if hocr_page_html:
            if self.mode == "IA":
                bboxes = self.get_boxes_for_IA(hocr_page_html)
            else:
                bboxes = self.get_boxes(hocr_page_html)
        return bboxes, None



from typing import Any, Dict, List, Optional

from openpecha.config import get_logger
from openpecha.exceptions import MetaDataMissingError, MetaDataValidationError
from openpecha.pecha import Pecha
from openpecha.pecha.annotations import AnnotationModel
from openpecha.pecha.pecha_types import PechaType, get_aligned_id, get_pecha_type
from openpecha.pecha.serializers.pecha_db.commentary.prealigned_commentary import (
    PreAlignedCommentarySerializer,
)
from openpecha.pecha.serializers.pecha_db.commentary.simple_commentary import (
    SimpleCommentarySerializer,
)
from openpecha.pecha.serializers.pecha_db.prealigned_root_translation import (
    PreAlignedRootTranslationSerializer,
)
from openpecha.pecha.serializers.pecha_db.root import RootSerializer

logger = get_logger(__name__)


def has_segmentation_annotation(ann_models: List[AnnotationModel]) -> bool:
    for ann_model in ann_models:
        path = ann_model.path
        if path.split("/")[1].startswith("segmentation"):
            return True
    return False


def get_segmentation_id(ann_models: List[AnnotationModel]) -> Optional[str]:
    for ann_model in ann_models:
        path = ann_model.path
        if path.split("/")[1].startswith("segmentation"):
            return path
    return None


# Handler functions for each PechaType
def _serialize_root_pecha(
    pechas, metadatas, annotations, pecha_category, annotation_path
):
    return RootSerializer().serialize(pechas[0], annotation_path, pecha_category)


def _serialize_root_translation_pecha(
    pechas, metadatas, annotations, pecha_category, annotation_path
):
    root_alignment_id = get_aligned_id(annotations[pechas[0].id], annotation_path)
    logger.info(f"Root Alignment ID: {root_alignment_id}")
    return RootSerializer().serialize(
        pechas[1],
        root_alignment_id,
        pecha_category,
        pechas[0],
        annotation_path,
    )


def _serialize_commentary_pecha(
    pechas, metadatas, annotations, pecha_category, annotation_path
):
    root_title = Serializer.get_root_en_title(metadatas, pechas)
    return SimpleCommentarySerializer().serialize(
        pechas[0], annotation_path, pecha_category, root_title
    )


def _serialize_commentary_translation_pecha(
    pechas, metadatas, annotations, pecha_category, annotation_path
):
    root_title = Serializer.get_root_en_title(metadatas, pechas)

    commentary_pecha = pechas[1]
    translation_pecha = pechas[0]

    commentary_alignment_id = get_aligned_id(
        annotations[translation_pecha.id], annotation_path
    )

    return SimpleCommentarySerializer().serialize(
        commentary_pecha,
        commentary_alignment_id,
        pecha_category,
        root_title,
        translation_pecha,
        annotation_path,
    )


def _serialize_prealigned_commentary_pecha(
    pechas, metadatas, annotations, pecha_category, annotation_path
):
    root_pecha = pechas[1]
    commentary_pecha = pechas[0]

    root_alignment_id = get_aligned_id(
        annotations[commentary_pecha.id], annotation_path
    )

    commentary_ann_models = annotations[commentary_pecha.id]
    if has_segmentation_annotation(commentary_ann_models):
        commentary_segmentation_id = get_segmentation_id(commentary_ann_models)
        logger.info(
            f"Serializing PreAligned Commentary Pecha with segmentation id: {commentary_segmentation_id}"
        )
        return PreAlignedCommentarySerializer().serialize(
            root_pecha,
            root_alignment_id,
            commentary_pecha,
            annotation_path,
            pecha_category,
            commentary_segmentation_id,
        )
    else:
        return PreAlignedCommentarySerializer().serialize(
            root_pecha,
            root_alignment_id,
            commentary_pecha,
            annotation_path,
            pecha_category,
        )


def _serialize_prealigned_root_translation_pecha(
    pechas, metadatas, annotations, pecha_category, annotation_path
):
    root_pecha = pechas[1]
    translation_pecha = pechas[0]
    root_alignment_id = get_aligned_id(
        annotations[translation_pecha.id], annotation_path
    )

    translation_ann_models = annotations[translation_pecha.id]
    if has_segmentation_annotation(translation_ann_models):
        translation_segmentation_id = get_segmentation_id(translation_ann_models)
        return PreAlignedRootTranslationSerializer().serialize(
            root_pecha,
            root_alignment_id,
            translation_pecha,
            annotation_path,
            pecha_category,
            translation_segmentation_id,
        )
    else:
        return PreAlignedRootTranslationSerializer().serialize(
            root_pecha,
            root_alignment_id,
            translation_pecha,
            annotation_path,
            pecha_category,
        )


# Registry mapping PechaType to handler function
PECHA_SERIALIZER_REGISTRY = {
    PechaType.root_pecha: _serialize_root_pecha,
    PechaType.root_translation_pecha: _serialize_root_translation_pecha,
    PechaType.commentary_pecha: _serialize_commentary_pecha,
    PechaType.commentary_translation_pecha: _serialize_commentary_translation_pecha,
    PechaType.prealigned_commentary_pecha: _serialize_prealigned_commentary_pecha,
    PechaType.prealigned_root_translation_pecha: _serialize_prealigned_root_translation_pecha,
}


class Serializer:
    @staticmethod
    def get_root_en_title(metadatas: List[Any], pechas: List[Pecha]) -> str:
        """
        Commentary Pecha serialized JSON should have the root English title.
        """
        logger.info(f"Metadatachain: {metadatas}")

        root_metadata = metadatas[-1]
        root_pecha = pechas[-1]

        en_title = root_metadata.title["en"]
        if not en_title:
            logger.error(f"English title is missing in Root Pecha {root_pecha.id}.")
            raise MetaDataMissingError(
                f"English title is missing in Root Pecha {root_pecha.id}."
            )
        return en_title

    def serialize(
        self,
        pechas: List[Pecha],
        metadatas: List[Any],
        annotations: Dict[str, List[AnnotationModel]],
        pecha_category: List[Dict[str, Dict[str, str]]],
        annotation_path: str,
    ):
        """
        Serialize a Pecha based on its type.
        """
        pecha = pechas[0]
        pecha_type = get_pecha_type(pechas, metadatas, annotations, annotation_path)
        logger.info(f"Serializing Pecha {pecha.id}, Type: {pecha_type}")
        handler = PECHA_SERIALIZER_REGISTRY.get(pecha_type)
        if not handler:
            raise ValueError(f"Unsupported pecha type: {pecha_type}")
        return handler(pechas, metadatas, annotations, pecha_category, annotation_path)



from typing import Any, Dict, List, Optional

from stam import AnnotationStore

from openpecha.config import get_logger
from openpecha.exceptions import (
    FileNotFoundError,
    MetaDataMissingError,
    StamAnnotationStoreLoadError,
)
from openpecha.pecha import Pecha, get_annotations_data
from openpecha.pecha.serializers.pecha_tools.utils import get_base_text

logger = get_logger(__name__)


def get_layer_name(metadatas: List[Dict[str, Any]]) -> str:
    """
    Get the layer name from the metadata.
    """
    for metadata in metadatas:
        if "annotations" in metadata:
            return metadata["annotations"][0]["layer_name"]
    raise MetaDataMissingError("Layer name is missing in the metadata.")


class TranslationSerializer:
    def get_annotations_from_layer(self, pecha: Pecha, layer_path: str):
        ann_store_path = pecha.pecha_path.parent.joinpath(layer_path)
        if not ann_store_path.exists():
            logger.error(f"The layer path {str(ann_store_path)} does not exist.")
            raise FileNotFoundError(
                f"[Error] The layer path '{str(ann_store_path)}' does not exist."
            )
        else:
            try:
                annotation_store = AnnotationStore(file=ann_store_path.as_posix())
                annotations = get_annotations_data(annotation_store)
                return annotations
            except Exception as e:
                logger.error(
                    f"Unable to load annotation store from layer path: {ann_store_path}. {str(e)}"
                )
                raise StamAnnotationStoreLoadError(
                    f"[Error] Error loading annotation store from layer path: {layer_path}. {str(e)}"
                )

    def serialize(
        self,
        pechas: List[Pecha],
        metadatas: List[Dict[str, Any]],
        pecha_category: List[Dict[str, object]],
        layer_name: Optional[str] = None,
    ) -> Dict:
        # Get content from root and translation pecha
        pecha = pechas[0]
        metadata = metadatas[0]
        if layer_name is None:
            layer_name = get_layer_name(metadatas)

        layer_path = pecha.get_segmentation_layer_path()
        base_text = get_base_text(pecha, layer_path)
        annotations = self.get_annotations_from_layer(pecha, layer_path)

        serialized_json = {
            "base_text": base_text,
            "annotations": annotations,
            "metadata": metadata,
            "pecha_category": pecha_category,
        }
        logger.info(f"Pecha {pecha.id} is serialized successfully.")
        return serialized_json



from pathlib import Path
from typing import Any, List
from unittest.mock import patch

from openpecha.pecha import Pecha
from openpecha.pecha.layer import AnnotationType
from openpecha.pecha.parsers.docx.annotation import DocxAnnotationParser


class DocxAnnotationUpdate:
    def __init__(self):
        self.parser = DocxAnnotationParser()

    def extract_layer_name(self, layer_path: str) -> str:
        return Path(layer_path).stem

    def extract_layer_id(self, layer_path: str) -> str:
        layer_name = self.extract_layer_name(layer_path)
        return layer_name.split("-")[-1]

    def extract_layer_enum(self, layer_path: str) -> AnnotationType:
        layer_name = self.extract_layer_name(layer_path)
        return AnnotationType(layer_name.split("-")[0])

    def update_annotation(
        self,
        pecha: Pecha,
        annotation_path: str,
        docx_file: Path,
        metadatas: List[Any],
    ) -> Pecha:
        type = self.extract_layer_enum(annotation_path)
        layer_id = self.extract_layer_id(annotation_path)

        with patch("openpecha.pecha.get_layer_id") as mock_layer_id:
            mock_layer_id.return_value = layer_id
            updated_pecha, _ = self.parser.add_annotation(
                pecha, type, docx_file, metadatas
            )

        return updated_pecha



import json
import re
from pathlib import Path
from typing import Dict, List

from openpecha.config import PECHAS_PATH
from openpecha.pecha import Pecha
from openpecha.pecha.layer import AnnotationType
from openpecha.pecha.parsers import BaseParser


class ChonjukChapterParser(BaseParser):
    def __init__(self):
        self.regex = (
            r"ch(\d+)-\"([\u0F00-\u0FFF]+)\"\s*([\u0F00-\u0FFF\s\n]+)[\u0F00-\u0FFF]"
        )
        self.updated_text = ""
        self.annotations: List[Dict] = []

    def get_initial_ann_spans(self, text: str):
        """
        Process: - Extract Chapter annotations in the text before removing the string annotations
                 - Store the spans of the chapter annotations
        Output: Return the initial chapter annotations
        """

        # Find all matches
        matches = re.finditer(self.regex, text)

        chapter_anns = []
        # Iterate over the matches and store the spans
        for match in matches:
            curr_match = {
                "chapter_number": match.span(1),
                "chapter_title": match.span(2),
                AnnotationType.CHAPTER.value: match.span(3),
            }
            chapter_anns.append(curr_match)
        return chapter_anns

    def get_updated_text(self, text: str):
        """
        Process: Remove the chapter string annotations from the text
        """
        pattern = r"ch\d+-\"[\u0F00-\u0FFF]+\"\s*"

        cleaned_text = re.sub(pattern, "", text)
        return cleaned_text

    def get_annotations(self, text: str):
        """
        Process: Update the chapter annotations after removing the string annotations
        Output: Return the updated chapter annotations
        """
        initial_ann_spans = self.get_initial_ann_spans(text)
        chapter_anns = []
        offset = 0
        for ann_spans in initial_ann_spans:
            start, end = ann_spans["chapter_number"]
            chapter_number = text[start:end]

            start, end = ann_spans["chapter_title"]
            chapter_title = text[start:end]

            start, end = ann_spans[AnnotationType.CHAPTER.value]
            # Update the offset (2 is char length of 'ch' before chapter number)
            offset += start - ann_spans["chapter_number"][0] + 2

            Chapter_span = start - offset, end - offset
            chapter_anns.append(
                {
                    "chapter_number": chapter_number,
                    "chapter_title": chapter_title,
                    AnnotationType.CHAPTER.value: {
                        "start": Chapter_span[0],
                        "end": Chapter_span[1],
                    },
                }
            )
        return chapter_anns

    def parse(
        self,
        input: str,
        metadata: Dict | Path,
        output_path: Path = PECHAS_PATH,
    ) -> Pecha:
        if isinstance(metadata, Path):
            with open(metadata) as f:
                metadata = json.load(f)
        assert isinstance(metadata, dict)

        output_path.mkdir(parents=True, exist_ok=True)

        self.cleaned_text = self.get_updated_text(input)
        self.annotations = self.get_annotations(input)

        pecha = Pecha.create(output_path)
        base_name = pecha.set_base(self.cleaned_text)
        layer, _ = pecha.add_layer(base_name, AnnotationType.CHAPTER)

        for ann in self.annotations:
            pecha.add_annotation(layer, ann, AnnotationType.CHAPTER)

        pecha.set_metadata({"id": pecha.id, "parser": self.name, **metadata})
        layer.save()
        return pecha



import json
import re
from pathlib import Path
from typing import Any, Dict, List

import pyewts

from openpecha.config import PECHAS_PATH
from openpecha.pecha import Pecha
from openpecha.pecha.layer import AnnotationType
from openpecha.pecha.parsers import BaseParser

ewts = pyewts.pyewts()


class DharamanexusParser(BaseParser):
    def __init__(self, regex_pattern):
        self.regex_pattern = regex_pattern
        self.state = {}
        self.temp_state: Dict[str, Any] = {
            "base_text": "",
            "annotations": {"segments": {}, "pages": {}},
            "prev_info_dict": {},
        }

    def get_category_wise_files(self, files_json_path):
        with open(files_json_path) as f:
            files = json.load(f)
        category_file_paths: Dict = {}
        for file in files:
            category = file["category"]
            text_name = file["textname"]
            if category not in category_file_paths.keys():
                category_file_paths[category] = {text_name: [file["filename"]]}
            else:
                if text_name not in category_file_paths[category].keys():
                    category_file_paths[category][text_name] = [file["filename"]]
                else:
                    category_file_paths[category][text_name].append(file["filename"])
        return category_file_paths

    def read_json(self, file_path):
        with open(file_path) as f:
            return json.load(f)

    def sort_file_names(self, files):
        parts = re.findall(self.regex_pattern, files)
        key: List[int | float] = []
        for part in parts:
            if part.isdigit():
                key.append(int(part))
            elif re.match(r"\d+\.\d+", part):
                key.append(float(part))
            else:
                key.append(part.strip())
        return key

    def get_info_dict(self, segment_id, page_id, text):
        text_len = len(text)
        curr_dict: Dict[str, Any] = {"annotations": {"segments": {}, "pages": {}}}
        if self.temp_state["prev_info_dict"]:
            segment_start = (
                self.temp_state["prev_info_dict"]["segments"]["span"]["end"] + 1
            )
            segment_end = segment_start + text_len
            curr_dict["annotations"]["segments"][segment_id] = {
                "span": {"start": segment_start, "end": segment_end}
            }
            if page_id == self.temp_state["prev_info_dict"]["page_id"]:
                page_start = self.temp_state["prev_info_dict"]["pages"]["span"]["start"]
                curr_dict["annotations"]["pages"][page_id] = {
                    "span": {"start": page_start, "end": segment_end}
                }
            elif page_id != self.temp_state["prev_info_dict"]["page_id"]:
                page_start = (
                    self.temp_state["prev_info_dict"]["pages"]["span"]["end"] + 1
                )
                curr_dict["annotations"]["pages"][page_id] = {
                    "span": {"start": page_start, "end": page_start + text_len}
                }
        else:
            curr_dict["annotations"]["segments"][segment_id] = {
                "span": {"start": 0, "end": text_len}
            }
            curr_dict["annotations"]["pages"][page_id] = {
                "span": {"start": 0, "end": text_len}
            }
        return curr_dict

    def get_temp_state(self, json_file: List[Dict[str, Any]]) -> None:
        for info in json_file:
            text = ewts.toUnicode(info["original"])
            segment_id = info["segmentnr"]
            page_id = info["folio"]
            curr_dict = self.get_info_dict(segment_id, page_id, text)
            if self.temp_state["base_text"] == "":
                self.temp_state["base_text"] = text
            elif (
                self.temp_state["base_text"] != ""
                and self.temp_state["prev_info_dict"]["page_id"] == page_id
            ):
                self.temp_state["base_text"] += " " + text
            elif self.temp_state["prev_info_dict"]["page_id"] != page_id:
                self.temp_state["base_text"] += "\n" + text
            self.temp_state["annotations"]["segments"][segment_id] = curr_dict[
                "annotations"
            ]["segments"][segment_id]
            self.temp_state["annotations"]["pages"][page_id] = curr_dict["annotations"][
                "pages"
            ][page_id]
            self.temp_state["prev_info_dict"] = {
                "segment_id": segment_id,
                "page_id": page_id,
                "segments": curr_dict["annotations"]["segments"][segment_id],
                "pages": curr_dict["annotations"]["pages"][page_id],
            }

    def make_state(self, input):
        for vol, file_paths in input.items():
            for file_path in file_paths:
                json_file = self.read_json(file_path)
                self.get_temp_state(json_file)
            self.state[vol] = {
                "base_text": self.temp_state["base_text"],
                "annotations": self.temp_state["annotations"],
            }
            self.temp_state = {
                "base_text": "",
                "annotations": {"segments": {}, "pages": {}},
                "prev_info_dict": {},
            }

    def write_to_pecha(self, pecha, metadata):
        bases = {}
        order = 1
        for vol, data in self.state.items():
            base_name = pecha.set_base(content=data["base_text"])

            segment, _ = pecha.add_layer(base_name, AnnotationType.SEGMENTATION)
            for segment_id, segment_span in data["annotations"]["segments"].items():
                segment_ann = {
                    AnnotationType.SEGMENTATION.value: segment_span["span"],
                    "segment_id": segment_id,
                }
                pecha.add_annotation(segment, segment_ann, AnnotationType.SEGMENTATION)
            segment.save()

            pagination, _ = pecha.add_layer(base_name, AnnotationType.PAGINATION)
            for page_id, page_ann in data["annotations"]["pages"].items():
                page_ann = {
                    AnnotationType.PAGINATION.value: page_ann["span"],
                    "folio": page_id,
                }
                pecha.add_annotation(pagination, page_ann, AnnotationType.PAGINATION)
            pagination.save()

            curr_base = {
                "source_metadata": {
                    "source_id": vol,
                    "total_segments": len(data["annotations"]["segments"]),
                    "total_pages": len(data["annotations"]["pages"]),
                },
                "base_file": base_name,
                "order": order,
            }

            bases[base_name] = curr_base
            order += 1

        pecha.set_metadata(
            {"id": pecha.id, "parser": self.name, "bases": bases, **metadata}
        )

    def parse(
        self,
        input: Any,
        metadata: Dict | Path,
        output_path: Path = PECHAS_PATH,
    ):
        self.make_state(input)

        pecha = Pecha.create(output_path)
        self.write_to_pecha(pecha, metadata)
        return pecha

    def get_sorted_file_paths(self, file_paths):
        for _, texts in file_paths.items():
            sorted_texts = sorted(texts.keys(), key=self.sort_file_names)
            for text_name in sorted_texts:
                files = texts[text_name]
                sorted_files = sorted(files, key=self.sort_file_names)
        return sorted_files



import json
import re
from enum import Enum
from typing import Dict, Optional

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    GetCoreSchemaHandler,
    field_validator,
    model_validator,
)
from pydantic_core import core_schema

from openpecha.ids import get_uuid
from openpecha.pecha.layer import AnnotationType


class Span(BaseModel):
    start: int = Field(..., ge=0)
    end: int = Field(..., ge=0)
    errors: Optional[Dict] = None

    model_config = ConfigDict(extra="forbid")

    @field_validator("start", "end")
    @classmethod
    def span_must_not_be_neg(cls, v: int) -> int:
        if v < 0:
            raise ValueError("span shouldn't be negative")
        return v

    @model_validator(mode="after")
    def end_must_not_be_less_than_start(self) -> "Span":
        if self.end < self.start:
            raise ValueError("Span end must not be less than start")
        return self


class AnnBase(BaseModel):
    span: Span
    metadata: Optional[Dict] = Field(default=None)

    model_config = ConfigDict(extra="forbid")


class Page(AnnBase):
    page_info: Optional[str] = Field(default=None, description="page payload")
    imgnum: Optional[int] = Field(
        default=None,
        description="image sequence no. from bdrc api, http://iiifpres.bdrc.io/il/v:bdr:I0888",
    )
    reference: Optional[str] = Field(
        default=None, description="image filename from bdrc"
    )


class BaseAnnotation(BaseModel):
    span: Span
    metadata: Optional[Dict] = None

    model_config = ConfigDict(extra="forbid")


class Pagination(BaseAnnotation):
    page_info: Optional[str] = Field(default=None, description="page payload")
    imgnum: Optional[int] = Field(default=None, description="image sequence number")
    order: Optional[int] = Field(default=None, description="order of the page")
    reference: Optional[str] = Field(
        default=None, description="can be url or just string indentifier of source page"
    )


class Lang(BaseAnnotation):
    language: Optional[str] = Field(
        default=None, description="BCP-47 tag of the language"
    )


class OCRConfidence(BaseAnnotation):
    confidence: float
    nb_below_threshold: Optional[int] = None


class Citation(BaseAnnotation):
    pass


def _get_annotation_class(layer_name: AnnotationType):
    """Maps AnnotationType to Annotation class"""

    if layer_name == AnnotationType.PAGINATION:
        return Pagination
    elif layer_name == AnnotationType.LANGUAGE:
        return Lang
    elif layer_name == AnnotationType.CITATION:
        return Citation
    elif layer_name == AnnotationType.OCR_CONFIDENCE:
        return OCRConfidence
    else:
        return BaseAnnotation


class Layer(BaseModel):
    id: str = Field(default=None)
    annotation_type: AnnotationType
    revision: str = Field(default="00001")
    annotations: Dict = Field(default_factory=dict)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_validator(mode="before")
    def set_id(cls, values):
        values["id"] = values.get("id") or get_uuid()
        return values

    @field_validator("revision")
    def revision_must_int_parsible(cls, v):
        if not v.isdigit():
            raise ValueError("revision must be integer-parsable like `00002`")
        return v

    def bump_revision(self):
        self.revision = f"{int(self.revision)+1:05}"  # noqa

    def reset(self):
        self.revision = "00001"
        self.annotations = {}

    def get_annotations(self):
        """Yield Annotation Objects"""
        for ann_id, ann_dict in self.annotations.items():
            ann_class = _get_annotation_class(self.annotation_type)
            ann = ann_class.model_validate(ann_dict)
            yield ann_id, ann

    def get_annotation(self, annotation_id: str) -> Optional[BaseAnnotation]:
        """Retrieve annotation of id `annotation_id`"""
        ann_dict = self.annotations.get(annotation_id)
        if not ann_dict:
            return None
        ann_class = _get_annotation_class(self.annotation_type)
        ann = ann_class.model_validate(ann_dict)
        return ann

    def set_annotation(self, ann: BaseAnnotation, ann_id=None):
        """Add or Update annotation `ann` to the layer, returns the annotation id"""
        ann_id = ann_id if ann_id is not None else get_uuid()
        self.annotations[ann_id] = json.loads(ann.model_dump_json())
        return ann_id

    def remove_annotation(self, annotation_id: str):
        """Delete annotaiton of `annotation_id` from the layer"""
        if annotation_id in self.annotations:
            del self.annotations[annotation_id]


class OCRConfidenceLayer(Layer):
    confidence_threshold: float
    annotation_type: AnnotationType = Field(default=AnnotationType.OCR_CONFIDENCE)


class PechaId(str):
    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: type, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.no_info_after_validator_function(
            cls.validate, core_schema.str_schema()
        )

    @classmethod
    def validate(cls, v, info=None):  # <-- add info=None
        if not re.fullmatch(r"^I[A-F0-9]{8}$", v):
            raise ValueError(
                "PechaId must start with 'I' followed by 8 uppercase hex characters"
            )
        return v


class PechaRelationship(str, Enum):
    commentary_of = "commentary_of"
    translation_of = "translation_of"


class PechaAlignment(BaseModel):
    pecha_id: PechaId = Field(..., description="Pecha ID")
    alignment_id: Optional[str] = Field(
        default=None, pattern="\\S", description="Alignment ID"
    )


class AnnotationModel(BaseModel):
    pecha_id: PechaId = Field(..., description="Pecha ID")
    type: AnnotationType = Field(
        AnnotationType.SEGMENTATION, description="Type of the annotation"
    )
    document_id: str = Field(..., pattern="\\S")
    path: str = Field(..., pattern="\\S")
    title: str = Field(..., min_length=1)
    aligned_to: PechaAlignment | None = Field(None, description="Alignment descriptor")

    model_config = ConfigDict(
        extra="forbid",
        json_schema_extra={
            "examples": {
                "pecha_id": "I857977C3",
                "type": "alignment",
                "document_id": "1vgnfCQH3yaWPDaMDFXT_5GhlG0M9kEra0mxkDX46VLE",
                "annotation_id": "test_id",
                "title": "Test Alignment",
                "aligned_to": {
                    "pecha_id": "I857977C3",
                    "alignment_id": "test_alignment_id",
                },
            }
        },
    )



from typing import Dict, List

from stam import AnnotationStore

from openpecha.config import get_logger
from openpecha.exceptions import FileNotFoundError, StamAnnotationStoreLoadError
from openpecha.pecha import Pecha
from openpecha.pecha.metadata import Language
from openpecha.pecha.serializers.pecha_db.utils import (
    FormatPechaCategory,
    get_metadata_for_pecha_org,
)
from openpecha.utils import chunk_strings

logger = get_logger(__name__)


class RootSerializer:
    @staticmethod
    def get_texts_from_layer(layer: AnnotationStore):
        """
        Extract texts from layer
        1.If text is a newline, replace it with empty string
        2.Replace newline with <br>
        """
        return [
            "" if str(ann) == "\n" else str(ann).replace("\n", "<br>") for ann in layer
        ]

    def get_root_content(self, pecha: Pecha, layer_path: str):
        ann_store_path = pecha.pecha_path.parent.joinpath(layer_path)
        if not ann_store_path.exists():
            logger.error(f"The layer path {str(ann_store_path)} does not exist.")
            raise FileNotFoundError(
                f"[Error] The layer path '{str(ann_store_path)}' does not exist."
            )

        try:
            segment_layer = AnnotationStore(file=ann_store_path.as_posix())
        except Exception as e:
            logger.error(
                f"Unable to load annotation store from layer path: {ann_store_path}. {str(e)}"
            )
            raise StamAnnotationStoreLoadError(
                f"[Error] Error loading annotation store from layer path: {layer_path}. {str(e)}"
            )
        else:
            segments = self.get_texts_from_layer(segment_layer)
            return segments

    def get_translation_content(self, pecha: Pecha, layer_path: str):
        """
        Processes:
        1. Get the first txt file from root and translation opf
        2. Read meaning layer from the base txt file from each opfs
        3. Read segment texts and fill it to 'content' attribute in json formats
        """
        ann_store_path = pecha.pecha_path.parent.joinpath(layer_path)
        if not ann_store_path.exists():
            logger.error(f"The layer path {str(ann_store_path)} does not exist.")
            raise FileNotFoundError(
                f"[Error] The layer path '{str(ann_store_path)}' does not exist."
            )

        try:
            translation_segment_layer = AnnotationStore(file=ann_store_path.as_posix())
        except Exception as e:
            logger.error(
                f"Unable to load annotation store from layer path: {ann_store_path}. {str(e)}"
            )
            raise StamAnnotationStoreLoadError(
                f"[Error] Error loading annotation store from layer path: {ann_store_path}. {str(e)}"
            )
        else:
            segments: Dict[int, List[str]] = {}
            for ann in translation_segment_layer:
                ann_data = {}
                for data in ann:
                    ann_data[str(data.key().id())] = data.value().get()

                if "root_idx_mapping" in ann_data:
                    root_map = int(ann_data["root_idx_mapping"])
                    segments[root_map] = [str(ann)]

            max_root_idx = max(segments.keys())
            translation_segments = []
            for root_idx in range(1, max_root_idx + 1):
                if root_idx in segments:
                    translation_segments.append("".join(segments[root_idx]))
                else:
                    translation_segments.append("")

            return translation_segments

    def serialize(
        self,
        pecha: Pecha,
        ann_id: str,
        pecha_category: List[Dict],
        translation_pecha: Pecha | None = None,
        translation_ann_id: str | None = None,
    ) -> Dict:

        # Format Category
        formatted_category = FormatPechaCategory().format_root_category(
            pecha, pecha_category
        )
        root_category, translation_category = (
            formatted_category["bo"],
            formatted_category["en"],
        )
        logger.info("Pecha Category successfully formatted.")
        logger.info(f"Root Category: {root_category}")
        logger.info(f"Root Translation Category: {translation_category}")

        # Get the metadata for root and translation pecha
        root_metadata = get_metadata_for_pecha_org(pecha)
        logger.info(f"Root metadata for pecha.org: {root_metadata} ")

        if translation_pecha:
            translation_metadata = get_metadata_for_pecha_org(translation_pecha)
        else:
            translation_metadata = get_metadata_for_pecha_org(
                pecha, lang=Language.english.value
            )
        logger.info(f"Root Translation metadata for pecha.org {translation_metadata}")

        # Get content from root and translation pecha
        root_content = self.get_root_content(pecha, pecha.layer_path / ann_id)
        if translation_pecha:
            translation_content = self.get_translation_content(
                translation_pecha, translation_pecha.layer_path / translation_ann_id
            )
        else:
            translation_content = []
        logger.info("Content successfully recieved from Pecha for serialization.")

        # Preprocess newlines in content
        root_content = [
            line.replace("\\n", "<br>").replace("\n", "<br>") for line in root_content
        ]
        translation_content = [
            line.replace("\\n", "<br>").replace("\n", "<br>")
            for line in translation_content
        ]

        # Chapterize content
        root_content = chunk_strings(root_content)
        translation_content = chunk_strings(translation_content)

        root_json: Dict[str, List] = {
            "categories": root_category,
            "books": [{**root_metadata, "content": root_content}],
        }

        translation_json = {
            "categories": translation_category,
            "books": [{**translation_metadata, "content": translation_content}],
        }

        # Set the content for source and target and set it to JSON
        serialized_json = {
            "source": translation_json,
            "target": root_json,
        }
        logger.info(f"Pecha {pecha.id} is serialized successfully.")
        return serialized_json



import shutil
import tempfile
from pathlib import Path

from openpecha.storages import (  # Adjust import based on your script name
    update_git_folder,
)


def test_update_git_folder():
    source = tempfile.mkdtemp(prefix="source_P0001")
    dest = tempfile.mkdtemp(prefix="dest_P0001")

    source_path = Path(source)
    dest_path = Path(dest)

    try:
        # Create test files in source directory
        (source_path / "base").mkdir(parents=True, exist_ok=True)
        (source_path / "base" / "0001.txt").write_text("Updated base Content!")
        (source_path / "layers" / "0001").mkdir(parents=True, exist_ok=True)
        (source_path / "layers" / "0001" / "Meaning_Segment-8679.json").write_text(
            "{Updated Annotation file}"
        )

        # Create test files in dest directory
        (dest_path / "base").mkdir(parents=True, exist_ok=True)
        (dest_path / "base" / "0001.txt").write_text("OLd base Content!")
        (dest_path / "layers" / "0001").mkdir(parents=True, exist_ok=True)
        (dest_path / "layers" / "0001" / "Meaning_Segment-8679.json").write_text(
            "{ Old Annotation file}"
        )

        # Create .git and .github directories that should not be deleted
        (dest_path / ".git").mkdir(exist_ok=True)
        (dest_path / ".github").mkdir(exist_ok=True)

        update_git_folder(source_path, dest_path)

        # Check if destination folder is updated correctly
        assert (dest_path / "base" / "0001.txt").exists()
        assert (dest_path / "layers" / "0001" / "Meaning_Segment-8679.json").exists()
        assert (dest_path / "base" / "0001.txt").read_text() == "Updated base Content!"
        assert (
            dest_path / "layers" / "0001" / "Meaning_Segment-8679.json"
        ).read_text() == "{Updated Annotation file}"

        # Ensure .git and .github directories are still present
        assert (dest_path / ".git").exists()
        assert (dest_path / ".github").exists()

    finally:
        shutil.rmtree(source)
        shutil.rmtree(dest)



from enum import Enum


class AnnotationCollectionType(Enum):
    """In STAM, this is used for setting DataSet id"""

    STRUCTURE_ANNOTATION = "structure_annotation"
    VARIATION_ANNOTATION = "variation_annotation"
    OCR_ANNOTATION = "ocr_annotation"
    LANGUAGE_ANNOTATION = "language_annotation"
    SEGMENTATION_ANNOTATION = "segmentation_annotation"


class AnnotationType(str, Enum):
    SEGMENTATION = "segmentation"
    ALIGNMENT = "alignment"

    CHAPTER = "chapter"
    PAGINATION = "pagination"
    DURCHEN = "durchen"
    SAPCHE = "sapche"

    OCR_CONFIDENCE = "ocr_confidence"
    LANGUAGE = "language"
    CITATION = "citation"
    BOOK_TITLE = "book_title"


class AnnotationGroupType(Enum):
    STRUCTURE_TYPE = "structure_type"
    SPELLING_VARIATION = "spelling_variation"
    OCR_CONFIDENCE_TYPE = "ocr_confidence_type"
    LANGUAGE_TYPE = "language_type"
    SEGMENTATION_TYPE = "segmentation_type"


def get_annotation_group_type(layer_type: AnnotationType) -> AnnotationGroupType:
    """return the annotation category where annotation type falls in"""

    if layer_type in [AnnotationType.SEGMENTATION, AnnotationType.ALIGNMENT]:
        return AnnotationGroupType.SEGMENTATION_TYPE

    if layer_type in [
        AnnotationType.CHAPTER,
        AnnotationType.SAPCHE,
        AnnotationType.PAGINATION,
    ]:
        return AnnotationGroupType.STRUCTURE_TYPE

    if layer_type == AnnotationType.LANGUAGE:
        return AnnotationGroupType.LANGUAGE_TYPE

    if layer_type == AnnotationType.OCR_CONFIDENCE:
        return AnnotationGroupType.OCR_CONFIDENCE_TYPE

    if layer_type == AnnotationType.DURCHEN:
        return AnnotationGroupType.SPELLING_VARIATION

    raise ValueError(f"Layer type {layer_type} has no defined AnnotationGroupType")


def get_annotation_collection_type(
    layer_type: AnnotationType,
) -> AnnotationCollectionType:
    """return the annotation category where annotation type falls in"""

    if layer_type in [AnnotationType.SEGMENTATION, AnnotationType.ALIGNMENT]:
        return AnnotationCollectionType.SEGMENTATION_ANNOTATION

    if layer_type in [
        AnnotationType.CHAPTER,
        AnnotationType.SAPCHE,
        AnnotationType.PAGINATION,
    ]:
        return AnnotationCollectionType.STRUCTURE_ANNOTATION

    if layer_type == AnnotationType.LANGUAGE:
        return AnnotationCollectionType.LANGUAGE_ANNOTATION

    if layer_type == AnnotationType.OCR_CONFIDENCE:
        return AnnotationCollectionType.OCR_ANNOTATION

    if layer_type == AnnotationType.DURCHEN:
        return AnnotationCollectionType.VARIATION_ANNOTATION

    raise ValueError(f"Layer type {layer_type} has no defined AnnotationCollectionType")



from pathlib import Path
from typing import Any, Dict, List, Tuple

from stam import AnnotationStore

from openpecha.config import get_logger
from openpecha.exceptions import ParseNotReadyForThisAnnotation
from openpecha.pecha import Pecha, annotation_path, get_anns
from openpecha.pecha.blupdate import DiffMatchPatch
from openpecha.pecha.layer import AnnotationType
from openpecha.pecha.parsers.docx.commentary.simple import DocxSimpleCommentaryParser
from openpecha.pecha.parsers.docx.root.number_list_root import DocxRootParser
from openpecha.pecha.pecha_types import is_root_related_pecha

pecha_id = str

logger = get_logger(__name__)


class DocxAnnotationParser:
    def __init__(self):
        pass

    def get_updated_coords(
        self, coords: List[Dict[str, int]], old_base: str, new_base: str
    ):
        diff_update = DiffMatchPatch(old_base, new_base)

        updated_coords = []
        for coord in coords:
            start = int(coord["start"])
            end = int(coord["end"])

            updated_coords.append(
                {
                    "start": diff_update.get_updated_coord(start),
                    "end": diff_update.get_updated_coord(end),
                    "root_idx_mapping": coord.get("root_idx_mapping", ""),
                }
            )

        return updated_coords

    def add_annotation(
        self,
        pecha: Pecha,
        type: AnnotationType | str,
        docx_file: Path,
        metadatas: List[Any],
    ) -> Tuple[Pecha, annotation_path]:

        # Accept both str and AnnotationType, convert str to AnnotationType
        if isinstance(type, str):
            try:
                type = AnnotationType(type)
            except ValueError:
                raise ParseNotReadyForThisAnnotation(f"Invalid annotation type: {type}")

        if type not in [AnnotationType.ALIGNMENT, AnnotationType.SEGMENTATION]:
            raise ParseNotReadyForThisAnnotation(
                f"Parser is not ready for the annotation type: {type}"
            )

        # New Segmentation Layer should be updated to this existing base
        new_basename = list(pecha.bases.keys())[0]
        new_base = pecha.get_base(new_basename)

        if is_root_related_pecha(metadatas):
            parser = DocxRootParser()
            coords, old_base = parser.extract_segmentation_coords(docx_file)

            updated_coords = self.get_updated_coords(coords, old_base, new_base)
            logger.info(f"Updated Coordinate: {updated_coords}")

            annotation_path = parser.add_segmentation_layer(pecha, updated_coords, type)
            anns = get_anns(
                AnnotationStore(file=str(pecha.layer_path / annotation_path))
            )
            logger.info(f"New Updated Annotations: {anns}")

            logger.info(
                f"Alignment Annotation is successfully added to Pecha {pecha.id}"
            )
            return (pecha, annotation_path)

        else:
            commentary_parser = DocxSimpleCommentaryParser()
            (
                coords,
                old_base,
            ) = commentary_parser.extract_segmentation_coords(docx_file)

            updated_coords = self.get_updated_coords(coords, old_base, new_base)
            annotation_path = commentary_parser.add_segmentation_layer(
                pecha, updated_coords, type
            )
            logger.info(
                f"Alignment Annotation is successfully added to Pecha {pecha.id}"
            )
            return (pecha, annotation_path)



from typing import Dict, List

from openpecha.alignment.commentary_transfer import CommentaryAlignmentTransfer
from openpecha.config import get_logger
from openpecha.pecha import Pecha
from openpecha.pecha.serializers.pecha_db.utils import (
    FormatPechaCategory,
    get_metadata_for_pecha_org,
    get_pecha_title,
)
from openpecha.utils import chunk_strings

logger = get_logger(__name__)


class PreAlignedCommentarySerializer:
    def serialize(
        self,
        root_pecha: Pecha,
        root_alignment_id: str,
        commentary_pecha: Pecha,
        commentary_alignment_id: str,
        pecha_category: List[Dict],
        commentary_segmentation_id: str | None = None,
    ):
        # Format Category
        root_title = get_pecha_title(root_pecha, "en")
        formatted_category = FormatPechaCategory().format_commentary_category(
            commentary_pecha, pecha_category, root_title
        )

        src_category, tgt_category = formatted_category["en"], formatted_category["bo"]
        logger.info(f"Category is extracted successfully for {commentary_pecha.id}.")

        # Get metadata
        src_metadata = get_metadata_for_pecha_org(commentary_pecha)
        tgt_metadata = get_metadata_for_pecha_org(commentary_pecha, "bo")
        logger.info(
            f"Pecha {commentary_pecha.id} Metadata successfully prepared for pecha.org website."
        )

        # Get content
        src_content: List[List[str]] = []
        if commentary_segmentation_id:
            tgt_content = (
                CommentaryAlignmentTransfer().get_serialized_commentary_segmentation(
                    root_pecha,
                    root_alignment_id,
                    commentary_pecha,
                    commentary_alignment_id,
                    commentary_segmentation_id,
                )
            )
        else:
            tgt_content = CommentaryAlignmentTransfer().get_serialized_commentary(
                root_pecha, root_alignment_id, commentary_pecha, commentary_alignment_id
            )
        logger.info(
            f"Alignment transfered content is extracted successfully for {commentary_pecha.id}."
        )
        # Preprocess newlines in content
        tgt_content = [
            line.replace("\\n", "<br>").replace("\n", "<br>") for line in tgt_content
        ]

        # Chapterize content
        chapterized_tgt_content = chunk_strings(tgt_content)

        serialized_json = {
            "source": {
                "categories": src_category,
                "books": [{**src_metadata, "content": src_content}],
            },
            "target": {
                "categories": tgt_category,
                "books": [{**tgt_metadata, "content": chapterized_tgt_content}],
            },
        }
        logger.info(f"Pecha {commentary_pecha.id} is serialized successfully.")
        return serialized_json



import enum
import os
import shutil
from pathlib import Path
from typing import Optional

import git
from git import GitCommandError, Repo
from github import Github

from openpecha.config import GITHUB_ORG_NAME
from openpecha.github_utils import create_github_repo

URL: str


class Storages(enum.Enum):
    GITHUB = enum.auto()


def _getenv(name, optional):
    env = os.getenv(name)
    if env:
        return env
    elif optional:
        return env
    else:
        raise RuntimeError(f"Please set {name} environment variable")


def _get_value(value, env_name, optional=False):
    return value if value else _getenv(env_name, optional)


def get_authenticated_remote_url(url, org, token):
    old_url = url.split("//")
    auth_remote_url = f"{old_url[0]}//{org}:{token}@{old_url[1]}"  # noqa
    return auth_remote_url


def setup_auth_for_new_repo(repo, org, token, remote_url):
    auth_remote_url = get_authenticated_remote_url(remote_url, org, token)
    repo.create_remote("origin", auth_remote_url)
    return repo


def is_repo_authenticated(repo) -> bool:
    remote_url = repo.remote().url
    if "@" in remote_url:
        return True
    return False


def setup_auth_for_old_repo(repo, org, token):
    remote_url = repo.remote().url
    if is_repo_authenticated(repo):
        return repo
    auth_remote_url = get_authenticated_remote_url(remote_url, org, token)
    repo.remote().set_url(auth_remote_url)
    return repo


def commit_and_push(repo, message, branch=None):
    try:
        if not branch:
            branch = repo.active_branch.name
        repo.git.add("-A")

        # Check if there are changes to commit
        if not repo.is_dirty(untracked_files=True):
            # "No changes to commit. Working tree is clean.
            return

        repo.git.commit("-m", message)
        repo.git.push("-u", "origin", branch)
    except GitCommandError as e:
        print(f"Git command error: {e}")


class Storage:
    def add_dir(
        self,
        path: Path,
        description: str,
        is_private: bool = False,
        branch: Optional[str] = None,
    ):
        raise NotImplementedError

    def remove_dir_with_name(self, name: str):
        raise NotImplementedError

    def remove_dir_with_path(self, path: Path):
        raise NotImplementedError


class GithubStorage(Storage):
    """class representing Github Storage

    This storage create, update, and delete github repo and files

    Args:
        org (str, optional): github organization name
        token (str, optional): github oauth token
        username (str, optional): github username
        email (str, optional): github linked email

    """

    def __init__(
        self,
        org: Optional[str] = None,
        token: Optional[str] = None,
        username: Optional[str] = None,
        email: Optional[str] = None,
    ):
        self.org_name = GITHUB_ORG_NAME
        self.token = _get_value(token, "GITHUB_TOKEN")
        self._username = _get_value(username, "GITHUB_USERNAME", optional=True)
        self._email = _get_value(email, "GITHUB_EMAIL", optional=True)
        self._org = None

    @property
    def org(self):
        if not self._org:
            self._org = Github(self.token).get_organization(self.org_name)
        return self._org

    @property
    def username(self):
        if not self._username:
            raise RuntimeError(
                "Please set usename attr or env variable GITHUB_USERNAME."
            )
        return self._username

    @property
    def email(self):
        if not self._email:
            raise RuntimeError("Please set email attr or env variable GITHUB_EMAIL.")
        return self._email

    def _init_local_repo(self, path: Path, remote_url: str):
        repo = Repo.init(path, initial_branch="main")
        repo = setup_auth_for_new_repo(repo, self.org_name, self.token, remote_url)
        repo.config_writer().set_value("user", "name", self.username).release()
        repo.config_writer().set_value("user", "email", self.email).release()
        return repo

    def _init_remote_repo(self, path: Path, description: str, is_private: bool):
        """Creates remote repo in Github and returns it's url."""
        remote_repo_url = create_github_repo(
            path=path,
            org_name=self.org_name,
            token=self.token,
            private=is_private,
            description=description,
        )
        return remote_repo_url

    def is_git_repo(self, path):
        try:
            _ = git.Repo(path).git_dir
            return True
        except git.exc.InvalidGitRepositoryError:
            return False

    def get_authenticated_repo_remote_url(self, repo_name: str):
        return f"https://{self.username}:{self.token}@github.com/{self.org_name}/{repo_name}.git"  # noqa

    def add_dir(
        self,
        path: Path,
        description: str,
        is_private: bool = False,
        branch: Optional[str] = None,
    ):
        """dir local dir to github."""
        remote_repo_url = self._init_remote_repo(
            path=path, description=description, is_private=is_private
        )
        local_repo = self._init_local_repo(path=path, remote_url=remote_repo_url)
        commit_and_push(repo=local_repo, message="Initial commit", branch=branch)
        return local_repo

    def remove_dir_with_name(self, name: str):
        repo = self.org.get_repo(name)
        repo.delete()

    def remove_dir_with_path(self, path: Path):
        """Remove repo with local path, assumes that local and remote name is same."""
        repo = self.org.get_repo(path.name)
        repo.delete()
        shutil.rmtree(str(path))

    def get_dir_with_name(self, name: str):
        repo = self.org.get_repo(name)
        return repo

    def get_dir_with_path(self, path: Path):
        repo = self.org.get_repo(path.name)
        return repo

    def add_file(
        self, dir_name: str, path: str, content: str, message: str, update=False
    ):
        """add file to `name` repo.

        Args:
            dir_name: name of repo to add file to.
            path: path to add the file.
            content: content of the file.
            message: git commit message.
        """

        repo = self.org.get_repo(dir_name)
        if update:
            old_content = repo.get_contents(path)
            repo.update_file(old_content.path, message, content, old_content.sha)
        else:
            repo.create_file(path, message, content)

    def get_file(self, dir_name, path: str, branch="master"):
        repo = self.org.get_repo(dir_name)
        contents = repo.get_contents(path, ref=branch)
        return contents

    def remove_file(
        self, dir_name: str, path: str, message: str, branch: str = "master"
    ):
        repo = self.org.get_repo(dir_name)
        contents = repo.get_contents(path, ref=branch)
        repo.delete_file(contents.path, message, contents.sha, branch=branch)


def update_git_folder(update_source_folder_path: Path, update_dest_folder_path: Path):
    # delete files from the new repo
    for file in update_dest_folder_path.glob("*"):
        if file.name in [".git", ".github"]:
            continue

        if file.is_dir():
            shutil.rmtree(file)
        else:
            file.unlink()

    # Copy files from input data dir to new git repo
    for file in update_source_folder_path.rglob("*"):
        if file.is_file():
            relative_path = file.relative_to(update_source_folder_path)
            dest_path = update_dest_folder_path / relative_path

            dest_path.parent.mkdir(parents=True, exist_ok=True)

            shutil.copy2(file, dest_path)


def update_github_repo(input_data_dir: Path, repo_path: Path):
    """
    Overwrite the files in the github repo with the files in the input data dir
    1. Clone the repo from the github organization
    2. Delete files from the new repo
    2. Copy files from input data dir to new git repo
    3. Commit and push the changes in main branch
    """

    update_git_folder(input_data_dir, repo_path)

    # Commit and push the changes in main branch
    local_repo = Repo(repo_path)
    # Get Git user details from environment variables
    git_user_name = os.getenv("GITHUB_USERNAME")
    git_user_email = os.getenv("GITHUB_EMAIL")

    with local_repo.config_writer() as config:
        config.set_value("user", "name", git_user_name)
        config.set_value("user", "email", git_user_email)

    commit_and_push(local_repo, message="Pecha update")



from pathlib import Path
from typing import Any, Dict, List

from stam import AnnotationStore

from openpecha.config import get_logger
from openpecha.pecha import Pecha, get_anns, load_layer
from openpecha.utils import adjust_segment_num_for_chapter, get_chapter_for_segment

logger = get_logger(__name__)


def is_empty(text: str) -> bool:
    """
    Return True if text is empty or contains only newlines.
    """
    return not text.strip().replace("\n", "")


def parse_root_mapping(mapping: str) -> List[int]:
    """
    Parse root_idx_mapping string like '1,2-4' into a sorted list of ints.
    """
    res = []
    for part in mapping.strip().split(","):
        part = part.strip()
        if "-" in part:
            start, end = part.split("-")
            res.extend(list(range(int(start), int(end) + 1)))
        else:
            res.append(int(part))
    res.sort()
    return res


class CommentaryAlignmentTransfer:
    @staticmethod
    def get_first_valid_root_idx(ann) -> int | None:
        indices = parse_root_mapping(ann["root_idx_mapping"])
        return indices[0] if indices else None

    @staticmethod
    def is_valid_ann(anns: Dict[int, Dict[str, Any]], idx: int) -> bool:
        return idx in anns and not is_empty(anns[idx]["text"])

    def get_segmentation_ann_path(self, pecha: Pecha) -> Path:
        """
        Return the path to the first segmentation layer JSON file in the pecha.
        """
        return next(pecha.layer_path.rglob("segmentation-*.json"))

    def index_annotations_by_root(
        self, anns: List[Dict[str, Any]]
    ) -> Dict[int, Dict[str, Any]]:
        """
        Return a dict mapping root_idx_mapping to the annotation dict.
        """
        return {int(ann["root_idx_mapping"]): ann for ann in anns}

    def map_layer_to_layer(
        self, src_layer: AnnotationStore, tgt_layer: AnnotationStore
    ) -> Dict[int, List[int]]:
        """
        Map annotations from src_layer to tgt_layer based on span overlap or containment.
        Returns a mapping from source indices to lists of target indices.
        """

        def extract_idx(ann: dict) -> int:
            """Helper to extract a single int index from root_idx_mapping."""
            if "-" in ann["root_idx_mapping"] or "," in ann["root_idx_mapping"]:
                idx = self.get_first_valid_root_idx(ann)
                if idx is None:
                    raise ValueError(
                        f"Invalid root_idx_mapping: {ann['root_idx_mapping']}"
                    )
                return idx
            return int(ann["root_idx_mapping"])

        def is_match(src_start, src_end, tgt_start, tgt_end):
            """Helper to check if spans overlap or are contained (not edge overlap)."""
            is_overlap = (
                src_start <= tgt_start < src_end or src_start < tgt_end <= src_end
            )
            is_contained = tgt_start < src_start and tgt_end > src_end
            is_edge_overlap = tgt_start == src_end or tgt_end == src_start
            return (is_overlap or is_contained) and not is_edge_overlap

        mapping: Dict[int, List[int]] = {}
        src_anns = get_anns(src_layer, include_span=True)
        tgt_anns = get_anns(tgt_layer, include_span=True)
        for src_ann in src_anns:
            src_start, src_end = src_ann["Span"]["start"], src_ann["Span"]["end"]
            try:
                src_idx = extract_idx(src_ann)
            except ValueError:
                continue
            mapping[src_idx] = []
            for tgt_ann in tgt_anns:
                tgt_start, tgt_end = tgt_ann["Span"]["start"], tgt_ann["Span"]["end"]
                try:
                    tgt_idx = extract_idx(tgt_ann)
                except ValueError:
                    continue
                if is_match(src_start, src_end, tgt_start, tgt_end):
                    mapping[src_idx].append(tgt_idx)
        return dict(sorted(mapping.items()))

    def get_root_pechas_mapping(
        self, pecha: Pecha, alignment_id: str
    ) -> Dict[int, List[int]]:
        """
        Get mapping from pecha's alignment layer to segmentation layer.
        """
        segmentation_ann_path = self.get_segmentation_ann_path(pecha)
        segmentation_layer = load_layer(segmentation_ann_path)
        alignment_layer = load_layer(pecha.layer_path / alignment_id)
        return self.map_layer_to_layer(alignment_layer, segmentation_layer)

    def get_commentary_pechas_mapping(
        self, pecha: Pecha, alignment_id: str, segmentation_id: str
    ) -> Dict[int, List[int]]:
        """
        Get mapping from pecha's segmentation layer to alignment layer.
        """
        segmentation_ann_path = pecha.layer_path / segmentation_id
        segmentation_layer = load_layer(segmentation_ann_path)
        alignment_layer = load_layer(pecha.layer_path / alignment_id)
        return self.map_layer_to_layer(segmentation_layer, alignment_layer)

    def get_serialized_commentary(
        self,
        root_pecha: Pecha,
        root_alignment_id: str,
        commentary_pecha: Pecha,
        commentary_alignment_id: str,
    ) -> List[str]:
        """
        Serialize commentary annotations with root/segmentation mapping and formatting.
        """
        root_map = self.get_root_pechas_mapping(root_pecha, root_alignment_id)
        root_segmentation_path = self.get_segmentation_ann_path(root_pecha)
        root_segmentation_anns = self.index_annotations_by_root(
            get_anns(load_layer(root_segmentation_path))
        )
        root_anns = self.index_annotations_by_root(
            get_anns(load_layer(root_pecha.layer_path / root_alignment_id))
        )
        commentary_anns = get_anns(
            load_layer(commentary_pecha.layer_path / commentary_alignment_id)
        )

        res: List[str] = []
        for ann in commentary_anns:
            result = self.process_commentary_ann(
                ann, root_anns, root_map, root_segmentation_anns
            )
            if result is not None:
                res.append(result)
        return res

    def get_serialized_commentary_segmentation(
        self,
        root_pecha: Pecha,
        root_alignment_id: str,
        commentary_pecha: Pecha,
        commentary_alignment_id: str,
        commentary_segmentation_id: str,
    ) -> List[str]:
        root_map = self.get_root_pechas_mapping(root_pecha, root_alignment_id)
        commentary_map = self.get_commentary_pechas_mapping(
            commentary_pecha, commentary_alignment_id, commentary_segmentation_id
        )

        root_segmentation_path = self.get_segmentation_ann_path(root_pecha)
        root_segmentation_anns = self.index_annotations_by_root(
            get_anns(load_layer(root_segmentation_path))
        )
        root_anns = self.index_annotations_by_root(
            get_anns(load_layer(root_pecha.layer_path / root_alignment_id))
        )
        commentary_segmentation_anns = get_anns(
            load_layer(commentary_pecha.layer_path / commentary_segmentation_id)
        )
        logger.info(
            "Root and Commentary Annotations retrieved for Commentary Transfer."
        )

        res: List[str] = []
        for ann in commentary_segmentation_anns:
            try:
                text = ann["text"]
                if is_empty(text):
                    continue

                if not commentary_map[int(ann["root_idx_mapping"])]:
                    res.append(text)
                    continue

                aligned_idx = commentary_map[int(ann["root_idx_mapping"])][0]
                if not self.is_valid_ann(root_anns, aligned_idx):
                    res.append(text)
                    continue

                if not root_map[aligned_idx]:
                    res.append(text)
                    continue

                root_display_idx = root_map[aligned_idx][0]
                if not self.is_valid_ann(root_segmentation_anns, root_display_idx):
                    res.append(text)
                    continue

                chapter_num = get_chapter_for_segment(root_display_idx)
                processed_root_display_idx = adjust_segment_num_for_chapter(
                    root_display_idx
                )
                res.append(
                    self.format_serialized_commentary(
                        chapter_num, processed_root_display_idx, text
                    )
                )
            except Exception as e:
                logger.error(
                    f"Error processing annotation: {ann}\nException: {e}", exc_info=True
                )
        return res

    @staticmethod
    def format_serialized_commentary(chapter_num: int, seg_idx: int, text: str) -> str:
        """Format the serialized commentary string."""
        return f"<{chapter_num}><{seg_idx}>{text}"

    def process_commentary_ann(
        self,
        ann: dict,
        root_anns: dict,
        root_map: dict,
        root_segmentation_anns: dict,
    ) -> str | None:
        """Process a single commentary annotation and return the serialized string, or None if not valid."""
        commentary_text = ann["text"]
        if is_empty(commentary_text):
            return None

        root_idx = self.get_first_valid_root_idx(ann)
        if root_idx is None or not self.is_valid_ann(root_anns, root_idx):
            return commentary_text

        if not root_map[root_idx]:
            return commentary_text

        root_display_idx = root_map[root_idx][0]
        if not self.is_valid_ann(root_segmentation_anns, root_display_idx):
            return commentary_text

        chapter_num = get_chapter_for_segment(root_display_idx)
        processed_root_display_idx = adjust_segment_num_for_chapter(root_display_idx)
        return self.format_serialized_commentary(
            chapter_num, processed_root_display_idx, commentary_text
        )



from typing import Dict, List

from openpecha.config import get_logger
from openpecha.exceptions import MetaDataMissingError
from openpecha.pecha import Pecha
from openpecha.pecha.metadata import Language
from openpecha.utils import get_text_direction_with_lang

logger = get_logger(__name__)


class FormatPechaCategory:
    def __init__(self):
        self.bo_root_category = {
            "name": "རྩ་བ།",
            "heDesc": "",
            "heShortDesc": "",
        }
        self.en_root_category = {
            "name": "Root text",
            "enDesc": "",
            "enShortDesc": "",
        }
        self.lzh_root_category = {
            "name": "根本颂",
            "heDesc": "",
            "heShortDesc": "",
        }
        self.bo_commentary_category = {
            "name": "འགྲེལ་བ།",
            "heDesc": "",
            "heShortDesc": "",
        }
        self.en_commentary_category = {
            "name": "Commentary text",
            "enDesc": "",
            "enShortDesc": "",
        }
        self.lzh_commentary_category = {
            "name": "注释",
            "heDesc": "",
            "heShortDesc": "",
        }

    def get_category(self, pecha_category: List[Dict[str, Dict]]):
        """
        Get the category of the Pecha
        """
        category: Dict = {}
        for cate_info in pecha_category:
            bo_name = cate_info["name"].get("bo", "") if cate_info["name"] else ""
            en_name = cate_info["name"].get("en", "") if cate_info["name"] else ""
            lzh_name = cate_info["name"].get("lzh", "") if cate_info["name"] else ""

            bo_desc = (
                cate_info["description"].get("bo", "")
                if cate_info["description"]
                else ""
            )
            en_desc = (
                cate_info["description"].get("en", "")
                if cate_info["description"]
                else ""
            )
            lzh_desc = (
                cate_info["description"].get("lzh", "")
                if cate_info["description"]
                else ""
            )
            bo_short_desc = (
                cate_info["short_description"].get("bo", "")
                if cate_info["short_description"]
                else ""
            )
            en_short_desc = (
                cate_info["short_description"].get("en", "")
                if cate_info["short_description"]
                else ""
            )
            lzh_short_desc = (
                cate_info["short_description"].get("lzh", "")
                if cate_info["short_description"]
                else ""
            )
            if category == {}:
                category = {
                    "bo": [],
                    "en": [],
                    "lzh": [],
                }
            category["bo"].append(
                {
                    "name": bo_name,
                    "heDesc": bo_desc,
                    "heShortDesc": bo_short_desc,
                }
            )
            category["en"].append(
                {
                    "name": en_name,
                    "enDesc": en_desc,
                    "enShortDesc": en_short_desc,
                }
            )
            category["lzh"].append(
                {
                    "name": lzh_name,
                    "heDesc": lzh_desc,
                    "heShortDesc": lzh_short_desc,
                }
            )

        return category

    def assign_category(self, category, type: str):
        if type == "root":
            category["bo"].append(self.bo_root_category)
            category["en"].append(self.en_root_category)
            category["lzh"].append(self.lzh_root_category)
        else:
            category["bo"].append(self.bo_commentary_category)
            category["en"].append(self.en_commentary_category)
            category["lzh"].append(self.lzh_commentary_category)
        return category

    def format_root_category(self, pecha: Pecha, pecha_category: List[Dict]):
        """
        1.Add Root section ie "རྩ་བ།" or "Root text" to category
        2.Add pecha title to category
        """
        category = self.get_category(pecha_category)
        logger.info(f"Pecha {pecha.id} category formatted to add pecha Title")

        bo_title = get_pecha_title(pecha, "bo")
        en_title = get_pecha_title(pecha, "en")
        lzh_title = get_pecha_title(pecha, "lzh") if Language.literal_chinese.value in pecha.metadata.title.keys() else ""
        logger.info(
            f"Pecha Title extracted. BO title: {bo_title}, EN title: {en_title}"
        )

        category = self.assign_category(category, "root")

        category["bo"].append({"name": bo_title, "heDesc": "", "heShortDesc": ""})
        category["en"].append({"name": en_title, "enDesc": "", "enShortDesc": ""})
        category["lzh"].append({"name": lzh_title, "heDesc": "", "heShortDesc": ""})

        return category

    def format_commentary_category(
        self,
        pecha: Pecha,
        pecha_category: List[Dict],
        root_title: str,
    ):
        """
        1.Add Commentary section ie "འགྲེལ་བ།" or "Commentary text" to category
        2.Add pecha title to category
        """
        category = self.get_category(pecha_category)

        bo_title = get_pecha_title(pecha, "bo")
        en_title = get_pecha_title(pecha, "en")
        lzh_title = get_pecha_title(pecha, "lzh") if Language.literal_chinese.value in pecha.metadata.title.keys() else ""
        category = self.assign_category(category, "commentary")

        category["bo"].append({"name": bo_title, "heDesc": "", "heShortDesc": ""})
        category["en"].append({"name": en_title, "enDesc": "", "enShortDesc": ""})
        category["lzh"].append({"name": lzh_title, "heDesc": "", "heShortDesc": ""})

        mapping = {
            "base_text_titles": [root_title],
            "base_text_mapping": "many_to_one",
            "link": "Commentary",
        }

        category["bo"][-1].update(mapping)
        category["en"][-1].update(mapping)
        category["lzh"][-1].update(mapping)

        return category


def get_metadata_for_pecha_org(pecha: Pecha, lang: str | None = None):
    """
    Extract required metadata from Pecha for `pecha.org` serialization
    """
    if not lang:
        lang = pecha.metadata.language.value
    direction = get_text_direction_with_lang(lang)

    title = (
        get_pecha_title(pecha, lang)
        if lang
        else get_pecha_title(pecha, pecha.metadata.language.value)
    )

    title = title if lang in ["bo", "en", "lzh"] else f"{title}[{lang}]"
    source = pecha.metadata.source if pecha.metadata.source else ""

    return {
        "title": title,
        "language": lang,
        "versionSource": source,
        "direction": direction,
        "completestatus": "done",
    }


def get_pecha_title(pecha: Pecha, lang: str):
    pecha_title = pecha.metadata.title

    if isinstance(pecha_title, dict):
        title = pecha_title.get(lang.lower()) or pecha_title.get(lang.upper())

    if title is None or title == "":
        logger.error(
            f"[Error] {lang.upper()} title not available inside metadata for {pecha.id} for Serialization."
        )
        raise MetaDataMissingError(
            f"[Error] {lang.upper()} title not available inside metadata for {pecha.id} for Serialization."
        )

    return title



from typing import Any, Dict, List

from pecha_org_tools.translation import (
    get_bo_content_translation,
    get_en_content_translation,
)

from openpecha.exceptions import MetaDataValidationError
from openpecha.pecha import Pecha
from openpecha.pecha.layer import AnnotationType
from openpecha.pecha.metadata import Language, PechaMetaData
from openpecha.utils import get_text_direction_with_lang


class ComplexCommentarySerializer:
    def extract_metadata(self, pecha: Pecha):
        """
        Extract neccessary metadata from opf for serialization to json
        """
        metadata: PechaMetaData = pecha.metadata

        if not isinstance(metadata.title, dict):
            raise MetaDataValidationError(
                f"[Error] Commentary Pecha {pecha.id} has no English or Tibetan Title."
            )

        source_title = metadata.title.get("en") or metadata.title.get("EN")
        target_title = metadata.title.get("bo") or metadata.title.get("BO")

        src_metadata = {
            "title": source_title,
            "language": "en",
            "versionSource": metadata.source if metadata.source else "",
            "direction": get_text_direction_with_lang("en"),
            "completestatus": "done",
        }

        tgt_metadata = {
            "title": target_title,
            "language": metadata.language.value if metadata.language else "bo",
            "versionSource": metadata.source if metadata.source else "",
            "direction": get_text_direction_with_lang(metadata.language),
            "completestatus": "done",
        }

        return src_metadata, tgt_metadata

    def add_root_reference_to_category(self, category: Dict[str, Any], root_title: str):
        """
        Modify the category format to the required format for pecha.org commentary
        """
        for lang in ["bo", "en"]:
            last_category = category[lang][-1]
            last_category.update(
                {
                    "base_text_titles": [root_title],
                    "base_text_mapping": "many_to_one",
                    "link": "Commentary",
                }
            )
        return category

    def get_sapche_anns(self, pecha: Pecha):
        """
        Get the sapche annotations from the sapche layer,
        """
        sapche_anns = []
        basename = next(pecha.base_path.rglob("*.txt")).stem
        sapche_layer, _ = pecha.get_layer_by_ann_type(basename, AnnotationType.SAPCHE)
        for ann in sapche_layer:
            start, end = ann.offset().begin().value(), ann.offset().end().value()
            # Get metadata of the annotation
            ann_metadata = {}
            for data in ann:
                ann_metadata[data.key().id()] = str(data.value())
            sapche_anns.append(
                {
                    "Span": {"start": start, "end": end},
                    "text": str(ann),
                    "sapche_number": ann_metadata["sapche_number"],
                }
            )

        return sapche_anns

    def get_meaning_segment_anns(self, pecha: Pecha):
        """
        Get the meaning segment annotations from the meaning segment layer,
        """
        meaning_segment_anns = []
        basename = next(pecha.base_path.rglob("*.txt")).stem
        meaning_segment_layer, _ = pecha.get_layer_by_ann_type(
            basename, AnnotationType.ALIGNMENT
        )
        for ann in meaning_segment_layer:
            start, end = ann.offset().begin().value(), ann.offset().end().value()
            # Get metadata of the annotation
            ann_metadata = {}
            for data in ann:
                ann_metadata[data.key().id()] = str(data.value())

            curr_ann = {
                "Span": {"start": start, "end": end},
                "text": str(ann),
            }

            if "root_idx_mapping" in ann_metadata:
                curr_ann["root_idx_mapping"] = ann_metadata["root_idx_mapping"]

            meaning_segment_anns.append(curr_ann)

        return meaning_segment_anns

    def get_content(self, pecha: Pecha):
        """
        Prepare content in the sapche annotations to the required format(Tree like structure)
        """

        def format_tree(tree):
            """
            Format sapche ann which is in tree like structure to desired format
            """
            formatted_tree = {}

            # Iterate over each key in the tree dictionary
            for key, value in tree.items():
                # Create a new dictionary for the current node with 'title' and 'data'
                formatted_tree[value["title"]] = {
                    "data": value["data"],
                }

                # If there are children, process each child and add them as separate keys
                for child_key, child_value in value["children"].items():
                    child_formatted = format_tree(
                        {child_key: child_value}
                    )  # Recursively format the child
                    formatted_tree[value["title"]].update(child_formatted)

            return formatted_tree

        sapche_anns = self.get_sapche_anns(pecha)
        self.get_text_related_to_sapche(pecha, sapche_anns)

        formatted_sapche_anns: Dict[str, Any] = {}

        for sapche_ann in sapche_anns:
            keys = sapche_ann["sapche_number"].strip(".").split(".")
            current = formatted_sapche_anns
            for key in keys:
                if key not in current:
                    current[key] = {
                        "children": {},
                        "title": sapche_ann["text"],
                        "data": sapche_ann["meaning_segments"],
                    }
                current = current[key]["children"]

        return format_tree(formatted_sapche_anns)

    @staticmethod
    def format_commentary_segment_ann(ann: Dict[str, Any], chapter_num: int = 1) -> str:
        """
        Format the commentary meaning segment annotation to the required format
        Input: ann: meaning segment annotation
        Required Format:
        <a><b>Text, where a is chapter number, b is root mapping number,
                    and Text is the meaning segment text

                    If root mapping number is not available, then just return the text
        Output Format: string
        """
        if "root_idx_mapping" in ann:
            return f"<{chapter_num}><{ann['root_idx_mapping']}>{ann['text'].strip()}"
        return ann["text"].strip()

    def get_text_related_to_sapche(self, pecha: Pecha, sapche_anns: List[Dict]):
        """
        Get the text related to the sapche annotations from meaning segment layer,
        and add to 'meaning_segments' key of sapche annotations
        """
        meaning_segment_anns = self.get_meaning_segment_anns(pecha)

        num_of_sapches = len(sapche_anns)
        for idx, sapche_ann in enumerate(sapche_anns):
            start = sapche_ann["Span"]["start"]
            end = sapche_ann["Span"]["end"]

            sapche_ann["meaning_segments"] = []

            # Determine the boundary for the next sapche annotation, if applicable
            next_start = (
                sapche_anns[idx + 1]["Span"]["start"]
                if idx < num_of_sapches - 1
                else None
            )

            for meaning_segment_ann in meaning_segment_anns:
                meaning_segment_start = meaning_segment_ann["Span"]["start"]
                meaning_segment_end = meaning_segment_ann["Span"]["end"]

                # Check if it's the last sapche annotation and include all meaning segments after it
                if next_start is None and meaning_segment_end >= end:
                    formatted_meaning_segment_ann = self.format_commentary_segment_ann(
                        meaning_segment_ann
                    )
                    sapche_ann["meaning_segments"].append(formatted_meaning_segment_ann)

                if next_start is None:
                    continue

                # Otherwise, include meaning segments between the current sapche and the next one
                if meaning_segment_start >= start and meaning_segment_end <= next_start:
                    formatted_meaning_segment_ann = self.format_commentary_segment_ann(
                        meaning_segment_ann
                    )
                    sapche_ann["meaning_segments"].append(formatted_meaning_segment_ann)

    def get_json_content(self, pecha: Pecha):
        """
        Fill the source and target content to the json format
        """
        content = self.get_content(pecha)

        bo_title = pecha.metadata.title.get("bo") or pecha.metadata.title.get("BO")

        pecha_lang = pecha.metadata.language

        if pecha_lang == Language.tibetan:
            pecha_lang = Language.english

        pecha_lang_lowercase = pecha_lang.value.lower()
        pecha_lang_uppercase = pecha_lang.value.upper()

        other_title = pecha.metadata.title.get(
            pecha_lang_lowercase
        ) or pecha.metadata.title.get(pecha_lang_uppercase)

        if pecha.metadata.language == Language.tibetan:
            src_content = {
                other_title: {
                    "data": [],
                    **get_en_content_translation(content),
                }
            }
            tgt_content = {bo_title: {"data": [], **content}}

        else:
            src_content = {other_title: {"data": [], **content}}
            tgt_content = {
                bo_title: {
                    "data": [],
                    **get_bo_content_translation(content),
                }
            }
        return (src_content, tgt_content)

    def serialize(
        self, pecha: Pecha, pecha_category: Dict[str, List[Dict]], root_title: str
    ):
        """
        Serialize the commentary pecha to json format
        """

        src_book, tgt_book = [], []
        src_metadata, tgt_metadata = self.extract_metadata(pecha)
        src_book.append(src_metadata)
        tgt_book.append(tgt_metadata)

        category = self.add_root_reference_to_category(pecha_category, root_title)
        src_category, tgt_category = category["en"], category["bo"]

        src_content, tgt_content = self.get_json_content(pecha)
        src_book[0]["content"] = src_content
        tgt_book[0]["content"] = tgt_content

        serialized_json = {
            "source": {"categories": src_category, "books": src_book},
            "target": {"categories": tgt_category, "books": tgt_book},
        }
        return serialized_json



import gzip
import json
import os
from pathlib import Path
from zipfile import ZipFile

from bs4 import BeautifulSoup

from openpecha.buda.api import get_image_list as buda_get_image_list
from openpecha.buda.api import image_group_to_folder_name
from openpecha.config import get_logger
from openpecha.utils import read_json

# Initialize the logger
logger = get_logger(__name__)


class GoogleVisionSource:
    def __init__(
        self,
        bdrc_scan_id,
        buda_data,
        ocr_import_info,
        ocr_disk_path,
    ):
        self.ocr_import_info = ocr_import_info
        self.ocr_disk_path = ocr_disk_path
        self.bdrc_scan_id = bdrc_scan_id
        self.buda_data = buda_data

    def get_image_list(self, image_group_id):
        """Get the image list from either local JSON file or BUDA API.

        Args:
            image_group_id (str): Image group ID of the volume

        Returns:
            list: List of image filenames
        """
        # Try to get image list from local file first
        if self.ocr_disk_path and os.path.exists(
            self.ocr_disk_path / f"{image_group_id}.json"
        ):
            buda_il = read_json(self.ocr_disk_path / f"{image_group_id}.json")
            if isinstance(buda_il, list):
                return list(map(lambda ii: ii["filename"], buda_il))
            else:
                logger.error(f"Expected list from read_json, got {type(buda_il)}")
                return []

        # If local file doesn't exist or path not provided, use BUDA API
        try:
            buda_il = buda_get_image_list(self.bdrc_scan_id, image_group_id)
            if isinstance(buda_il, list):
                return list(map(lambda ii: ii["filename"], buda_il))
            else:
                logger.error(
                    f"Expected list from buda_get_image_list, got {type(buda_il)}"
                )
                return []
        except ImportError:
            logger.error("BUDA API module not available")
            return []
        except Exception as e:
            logger.error(f"Error getting image list from BUDA API: {e}")
            return []

    def get_source_info(self):
        return self.buda_data

    def get_image_data(self, image_group_id, image_id):
        vol_folder = image_group_to_folder_name(self.bdrc_scan_id, image_group_id)
        # Fix for 'dict' object has no attribute 'rfind'
        if isinstance(image_id, dict) and "filename" in image_id:
            image_filename = image_id["filename"]
        else:
            image_filename = image_id

        expected_ocr_filename = image_filename[: image_filename.rfind(".")] + ".json.gz"
        image_ocr_path = self.ocr_disk_path / vol_folder / expected_ocr_filename
        ocr_object = None
        try:
            ocr_object = json.load(gzip.open(str(image_ocr_path), "rb"))
        except Exception as e:
            logger.exception(f"could not read {image_ocr_path}: {e}")
        return ocr_object


class BDRCGBSource:
    def __init__(
        self,
        bdrc_scan_id,
        buda_data,
        ocr_import_info,
        ocr_disk_path,
    ):
        self.ocr_import_info = ocr_import_info
        self.ocr_disk_path = ocr_disk_path
        self.bdrc_scan_id = bdrc_scan_id
        self.buda_data = buda_data
        self.image_info = {}
        self.cur_zip = None
        self.cur_image_group_id = None

    def _get_image_list(self, image_group_id):
        """Get the image list from either local JSON file or BUDA API.

        Args:
            image_group_id (str): Image group ID of the volume

        Returns:
            list: List of image filenames
        """
        # Try to get image list from local file first
        if self.ocr_disk_path and os.path.exists(
            self.ocr_disk_path / f"{image_group_id}.json"
        ):
            return read_json(self.ocr_disk_path / f"{image_group_id}.json")

        # If local file doesn't exist or path not provided, use BUDA API
        try:
            return buda_get_image_list(self.bdrc_scan_id, image_group_id)
        except ImportError:
            logger.error("BUDA API module not available")
            return []
        except Exception as e:
            logger.error(f"Error getting image list from BUDA API: {e}")
            return []

    def get_image_list(self, image_group_id):
        self.get_images_info(image_group_id)
        buda_il = self._get_image_list(image_group_id)
        # format should be a list of image_id (/ file names)
        if isinstance(buda_il, list):
            return list(map(lambda ii: ii["filename"], buda_il))
        else:
            logger.error(f"Expected list from _get_image_list, got {type(buda_il)}")
            return []

    def get_hocr_filename(self, image_id):
        if not self.images_info:
            logger.error(
                "images_info is None or empty. Did you forget to call get_images_info()?"
            )
            return None

        for filename, img_ref in self.images_info.items():
            if img_ref == image_id:
                return filename

        logger.warning(f"No matching filename found for image_id: {image_id}")
        return None

    def get_images_info(self, image_group_id):
        vol_folder = image_group_to_folder_name(self.bdrc_scan_id, image_group_id)
        image_info_path = (
            Path(f"{self.ocr_disk_path}") / "info" / vol_folder / "gb-bdrc-map.json"
        )
        self.images_info = read_json(image_info_path)

    def get_source_info(self):
        return self.buda_data

    def get_image_group_data(self, image_group_id):
        if image_group_id == self.cur_image_group_id and self.cur_zip is not None:
            return self.cur_zip
        vol_folder = image_group_to_folder_name(self.bdrc_scan_id, image_group_id)
        zip_path = Path(f"{self.ocr_disk_path}") / "output" / vol_folder / "html.zip"
        self.cur_zip = ZipFile(zip_path)
        return self.cur_zip

    def get_image_data(self, image_group_id, image_filename):
        hocr_filename = self.get_hocr_filename(image_filename) + ".html"
        zf = self.get_image_group_data(image_group_id)
        try:
            for filename in zf.filelist:
                if filename.filename.split("/")[-1] == hocr_filename:
                    with zf.open(filename.filename) as hocr_file:
                        return hocr_file.read()
        except KeyError:
            logger.error(f"Error: {image_filename} not found in image_info.")
            return None
        except Exception as e:
            logger.error(f"An unexpected error occurred: {e}")
            return None


class HOCRIASource:
    def __init__(
        self,
        bdrc_scan_id,
        buda_data,
        ocr_import_info,
        ocr_disk_path,
    ):
        self.ocr_import_info = ocr_import_info
        self.ocr_disk_path = ocr_disk_path
        self.bdrc_scan_id = bdrc_scan_id
        self.buda_data = buda_data
        self.image_hocr = {}
        self.image_info = {}

    def get_image_list(self, image_group_id):
        """Get image list for the given image group ID.

        This method can handle both local JSON files and BUDA API calls.

        Args:
            image_group_id (str): Image group ID of the volume

        Returns:
            list: List of image filenames
        """
        self.get_images_info(image_group_id)
        buda_il = self._get_image_list(image_group_id)
        # format should be a list of image_id (/ file names)
        if isinstance(buda_il, list):
            return list(map(lambda ii: ii["filename"], buda_il))
        else:
            logger.error(f"Expected list from _get_image_list, got {type(buda_il)}")
            return []

    def _get_image_list(self, image_group_id):
        """Get the raw image list data from either local file or BUDA API.

        Args:
            image_group_id (str): Image group ID of the volume

        Returns:
            list: Raw image list data
        """
        # Try to get image list from local file first
        if self.ocr_disk_path and os.path.exists(
            self.ocr_disk_path / f"{image_group_id}.json"
        ):
            return read_json(self.ocr_disk_path / f"{image_group_id}.json")

        # If local file doesn't exist or path not provided, use BUDA API
        try:
            return buda_get_image_list(self.bdrc_scan_id, image_group_id)
        except ImportError:
            logger.error("BUDA API module not available")
            return []
        except Exception as e:
            logger.error(f"Error getting image list from BUDA API: {e}")
            return []

    def get_images_info(self, image_group_id):
        curr_image = {}
        image_list = self._get_image_list(image_group_id)
        image_group_hocr = self.get_image_group_hocr(image_group_id)
        if image_group_hocr:
            hocr_html = BeautifulSoup(image_group_hocr, "html.parser")
            pages_hocr = hocr_html.find_all("div", {"class": "ocr_page"})
            for image_number, image_filename in enumerate(image_list):
                filename = image_filename["filename"]
                for page_hocr in pages_hocr:
                    if int(page_hocr["id"][5:]) == image_number:
                        curr_image[filename] = {"page_info": page_hocr}
                        self.image_info.update(curr_image)
                        curr_image = {}

    def get_image_group_hocr(self, image_group_id):
        vol_num = self.buda_data["image_groups"][image_group_id]["volume_number"]
        image_group_hocr_path = (
            Path(f"{self.ocr_disk_path}")
            / f"bdrc-{self.bdrc_scan_id}-{vol_num}_hocr.html"
        )
        try:
            hocr_html = image_group_hocr_path.read_text(encoding="utf-8")
            return hocr_html
        except FileNotFoundError:
            logger.error(f"Error: {image_group_hocr_path} not found.")
            return None
        except Exception as e:
            logger.error(f"An unexpected error occurred: {e}")
            return None

    def get_source_info(self):
        return self.buda_data

    def get_image_data(self, image_group_id, image_filename):
        try:
            page_hocr = self.image_info[image_filename]["page_info"]
            return page_hocr
        except KeyError:
            logger.error(f"Error: {image_filename} not found in image_info.")
            return None
        except Exception as e:
            logger.error(f"An unexpected error occurred: {e}")
            return None
